import torch
from secmlt.models.data_processing.data_processing import DataProcessing
from torchvision.transforms import Resize, Compose, Normalize, PILToTensor, ConvertImageDtype, Grayscale
from maltorch.datasets.grayscale_dataset import get_size
from PIL import Image


class GrayscalePreprocessing(DataProcessing):
    """
    Convert image to grayscale.
    """

    def __init__(self, width=256, height=256, convert_to_3d_image=True):
        super().__init__()
        self.width = width
        self.height = height
        self.convert_to_3d_image = convert_to_3d_image

        if self.convert_to_3d_image:
            self.preprocess = Compose(
                [
                    PILToTensor(),
                    ConvertImageDtype(torch.float),
                    Resize((self.width, self.height)),
                    Normalize(mean=0.5, std=0.5),
                    Grayscale(num_output_channels=3)
                ]
            )
        else:
            self.preprocess = Compose(
                [
                    PILToTensor(),
                    ConvertImageDtype(torch.float),
                    Resize((self.width, self.height)),
                    Normalize((0.5), (0.5))
                ]
            )

    def _process(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        x_flat = x.view(-1).tolist()
        width, height = get_size(len(x_flat))

        image = Image.new("L", (width, height))
        image.putdata(x_flat)
        image = self.preprocess(image)
        image = torch.cat([image, image, image])
        return torch.unsqueeze(image, 0)

    def invert(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        """
        Apply the inverted transform (if defined).

        Parameters
        ----------
        x : torch.Tensor
            Input samples.

        Returns
        -------
        torch.Tensor
            The samples in the input space before the transformation.
        """
        raise NotImplementedError



import torch
from secmlt.models.data_processing.data_processing import DataProcessing


class DeRandomizedPreprocessing(DataProcessing):
    """
    Daniel Gibert, Giulio Zizzo, Quan Le
    Certified Robustness of Static Deep Learning-based Malware Detectors against Patch and Append Attacks
    AISec '23

    Shoumik Saha, Weinxio Wang, Yigitcan Kaya, Soheil Feizi, Tudo Dumitras
    DRSM: De-Randomized Smoothing on Malware Classifier Providing Certified Robustness
    ICRL'24
    """
    def __init__(self, chunk_size: int = 512, padding_idx: int = 256):
        super().__init__()
        self.chunk_size = chunk_size
        self.padding_idx = padding_idx

    def _process(self, x: torch.Tensor) -> torch.Tensor:
        x = x.squeeze()  # Remove all dimensions equal to 1
        x = [x[i:i + self.chunk_size] for i in range(0, len(x), self.chunk_size)]
        x = torch.nn.utils.rnn.pad_sequence(x, batch_first=True, padding_value=self.padding_idx)
        return x

    def invert(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the inverted transform (if defined).

        Parameters
        ----------
        x : torch.Tensor
            Input samples.

        Returns
        -------
        torch.Tensor
            The samples in the input space before the transformation.
        """
        return x.view(-1)




"""
Daniel Gibert, Carles Mateu, Jordi Planes, Daniel Solis, Ramon Vicens
"Convolutional neural networks for classification of malware assembly code"
Recent Advances in Artificial Intelligence Research and Development: Proceedings of the 20th International Conference
of the Catalan Association for Artificial Intelligence.

Niall McLaughlin, Jesús Martínez del Rincón, BooJoong Kang, Suleiman Y. Yerima, Paul Miller, Sakir Sezer,
Yeganeh Safaei, Erik Trickel, Ziming Zhao, Adam Doupé, Gail-Joon Ahn
Deep Android Malware Detection.
ACM Conference on Data and Application Security and Privacy (CODASPY 2017)

Loosely based on the shallow convolutional neural network architecture defined in Gibert et al. 2017 and
McLaughlin et al. 2017 but with only a convolutional layer with kernel size equals to 3.
"""

import torch
import torch.nn.functional as F
from torch import nn
from maltorch.zoo.model import EmbeddingModel


class NGramConv(EmbeddingModel):
    def __init__(self,
                 embedding_size=8,
                 max_len=2 ** 20,
                 out_channels: int = 100,
                 threshold: float = 0.5,
                 padding_idx: int = 256):
        super(NGramConv, self).__init__(
            name="NGramConv", gdrive_id=None
        )
        self.embedding_1 = nn.Embedding(
            num_embeddings=257, embedding_dim=embedding_size, padding_idx=padding_idx
        )
        self.out_channels = out_channels
        self.conv1d_1 = nn.Conv1d(
            in_channels=embedding_size,
            out_channels=self.out_channels,
            kernel_size=(7,),
            stride=(3,),
            groups=1,
            bias=True,
        )

        self.dense_1 = nn.Linear(in_features=self.out_channels, out_features=self.out_channels, bias=True)
        self.drop_1 = nn.Dropout1d()
        self.dense_2 = nn.Linear(in_features=self.out_channels, out_features=1, bias=True)

        self.embedding_size = (embedding_size,)
        self.max_len = max_len
        self.threshold = threshold
        self.invalid_value = padding_idx
        self._expansion = torch.tensor([[-1.0, 1.0]])

    def embedding_layer(self):
        return self.embedding_1

    def embed(self, x):
        emb_x = self.embedding_1(x)
        emb_x = emb_x.transpose(1, 2)
        return emb_x

    def _forward_embed_x(self, x):
        conv1d_1 = self.conv1d_1(x)
        conv1d_1_activation = torch.relu(conv1d_1)

        global_max_pooling1d_1 = F.max_pool1d(
            input=conv1d_1_activation, kernel_size=conv1d_1_activation.size()[2:]
        )
        global_max_pooling1d_1_flatten = global_max_pooling1d_1.view(
            global_max_pooling1d_1.size(0), -1
        )

        dense_1 = self.dense_1(global_max_pooling1d_1_flatten)
        dense_1_activation = torch.relu(dense_1)
        drop_1 = self.drop_1(dense_1_activation)
        dense_2 = self.dense_2(drop_1)
        return dense_2

    def embedding_matrix(self):
        return self.embedding_1.weight



"""
Quo Vadis: Hybrid Machine Learning Meta-Model Based on Contextual and Behavioral Malware Representations
Dmitrijs Trizna, ACM AISec 2022
https://arxiv.org/abs/2208.12248
Reimplemented from: https://github.com/dtrizna/quo.vadis/blob/main/models.py
"""

import importlib.util
from typing import Dict, List, Union, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from secmlt.models.base_trainer import BaseTrainer
from secmlt.models.data_processing.data_processing import DataProcessing

from maltorch.zoo.malconv import MalConv
from maltorch.zoo.model import EmbeddingModel, BaseEmbeddingPytorchClassifier


class MLP(nn.Module):
    def __init__(
        self, 
        input_size: int, 
        output_size: int, 
        hidden_neurons: List[int] = None, 
        batch_norm: bool = False, 
        dropout: float = 0.0
    ):
        super().__init__()
        layers = []
        in_size = input_size

        # If hidden_neurons is None or empty, create a single layer mapping input to output
        if not hidden_neurons:
            layers.append(nn.Linear(in_size, output_size))
        else:
            # Add hidden layers
            for h in hidden_neurons:
                layers.append(nn.Linear(in_size, h))
                if batch_norm:
                    layers.append(nn.BatchNorm1d(h))
                layers.append(nn.ReLU())
                if dropout > 0:
                    layers.append(nn.Dropout(dropout))
                in_size = h
            # Add output layer
            layers.append(nn.Linear(in_size, output_size))

        self.net = nn.Sequential(*layers)
        
    def forward(self, x: torch.Tensor):
        return self.net(x)


class Core1DConvNet(EmbeddingModel):
    def __init__(
        self,
        name: str,
        gdrive_id: str,
        vocab_size: int = 152,
        embedding_size: int = 96,
            filter_sizes=None,
            num_filters=None,
        batch_norm_conv: bool = False,
        hidden_neurons: List[int] = [128],
        batch_norm_ffnn: bool = False,
        dropout: float = 0.5,
        output_dim: int = 128
    ):
        super().__init__(name, gdrive_id)
        if num_filters is None:
            num_filters = [128, 128, 128, 128]
        if filter_sizes is None:
            filter_sizes = [2, 3, 4, 5]
        self.embedding = nn.Embedding(
            num_embeddings=vocab_size,
            embedding_dim=embedding_size,
            padding_idx=0
        )

        self.conv1d_modules = nn.ModuleList()
        for filter_size, num_filter in zip(filter_sizes, num_filters):
            layers = [nn.Conv1d(
                in_channels=embedding_size,
                out_channels=num_filter,
                kernel_size=filter_size
            )]
            if batch_norm_conv:
                layers.append(nn.BatchNorm1d(num_filter))
            self.conv1d_modules.append(nn.Sequential(*layers))

        conv_output_size = sum(num_filters)

        self.mlp = MLP(
            input_size=conv_output_size,
            output_size=output_dim,
            hidden_neurons=hidden_neurons,
            batch_norm=batch_norm_ffnn,
            dropout=dropout
        )

    def embedding_layer(self):
        return self.embedding
    
    def embedding_matrix(self):
        return self.embedding.weight

    def embed(self, x: torch.Tensor):
        emb_x = self.embedding(x)
        emb_x = emb_x.permute(0, 2, 1)
        return emb_x
    
    def _forward_embed_x(self, x: torch.Tensor):
        conv_outputs = [
            self.conv_and_max_pool(x, conv_module)
            for conv_module in self.conv1d_modules
        ]
        x = torch.cat(conv_outputs, dim=1)  # [batch_size, conv_output_size]
        out = self.mlp(x)
        return out
     
    @staticmethod
    def conv_and_max_pool(x, conv_module: nn.Module):
        """Apply convolution and global max pooling."""
        x = conv_module(x)
        x = F.relu(x)
        x = x.max(dim=2)[0]  # Global max pooling over the sequence length
        return x

    def forward(self, inputs: torch.Tensor):
        # inputs: [batch_size, sequence_length]
        embedded = self.embedding(inputs)  # [batch_size, sequence_length, embedding_dim]
        embedded = embedded.permute(0, 2, 1)  # [batch_size, embedding_dim, sequence_length]
        
        # Apply convolution and pooling
        conv_outputs = [
            self.conv_and_max_pool(embedded, conv_module)
            for conv_module in self.conv1d_modules
        ]
        x = torch.cat(conv_outputs, dim=1)  # [batch_size, conv_output_size]
        out = self.mlp(x)
        return out


class Emulation(Core1DConvNet):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        speakeasy_spec = importlib.util.find_spec("speakeasy")
        if speakeasy_spec is None:
            raise ImportError("[-] 'speakeasy' is required for emulation module. Use 'pip install speakeasy' to install.")
        from maltorch.data_processing.dynamic.emulation import PEDynamicFeatureExtractor
        self.dynamic_extractor = PEDynamicFeatureExtractor(speakeasy_record_fields=['apis.api_name'], record_limits=None)

        nltk_spec = importlib.util.find_spec("nltk")
        if nltk_spec is None:
            raise ImportError("[-] 'nltk' is required for tokenization module. Use 'pip install nltk' to install.")
        from maltorch.data_processing.dynamic.tokenization import JSONTokenizerNaive
        # TODO: right now reusing APIs from Quo.Vadis paper
        # should be based on data in training set for fair comparison
        # so use: self.tokenizer.train(corpus, vocab_size)
        self.tokenizer = JSONTokenizerNaive(vocab="quovadis") 
 
    def emulate(self, pe: Union[str, bytes]):
        return self.dynamic_extractor.emulate(pe)
    
    def encode(self, report: dict):
        return self.tokenizer.encode(report)
    
    def preprocess(self, pe: Union[str, bytes]):
        report = self.emulate(pe)
        return self.encode(report)

    def forward(self, x: torch.Tensor):
        x = self.embed(x)
        output = self._forward_embed_x(x)
        return output
        
class QuoVadis(EmbeddingModel):
    def __init__(
            self,
            embedding_size: int = 8,
            padding_value: int = 256,
            representation_size: int = 128,
            quo_vadis_hidden_neurons: List[int] = [128],
            module_names=None,
    ):
        super(QuoVadis, self).__init__(
            name="QuoVadis", gdrive_id="NotYetPreTrained"
        )
        if module_names is None:
            module_names = ['malconv', 'emulation']
        self.embedding_size = embedding_size
        self.padding_value = padding_value
        self.module_names = module_names
        self.representation_size = representation_size
        self.representation_size_total = representation_size * len(self.module_set)
        self.module_set = self.init_modules()

        self.quovadis_meta_model = MLP(
            input_size=self.representation_size_total,
            output_size=1,
            hidden_neurons=quo_vadis_hidden_neurons,
            batch_norm=False,
            dropout=0.5
        )

    def init_modules(self) -> List[EmbeddingModel]:
        module_set = []
        for module in self.module_names:
            if module == "malconv":
                module_set.append(MalConv(
                    embedding_size=self.embedding_size,
                    padding_value=self.padding_value,
                    out_size=self.representation_size
                ))
            elif module == "emulation":
                module_set.append(Emulation(
                    embedding_size=self.embedding_size,
                    padding_value=self.padding_value,
                    output_dim=self.representation_size
                ))
            else:
                raise ValueError(f"Unknown module: {module}")
        return module_set

    @classmethod
    def create_model(
            cls,
            model_path: str = None,
            module_paths: Dict[str, str] = None,
            device: str = "cpu",
            preprocessing: DataProcessing = None,
            postprocessing: DataProcessing = None,
            trainer: BaseTrainer = None,
            threshold: Optional[float] = 0.5,
            **kwargs,
    ) -> BaseEmbeddingPytorchClassifier:
        net = cls(**kwargs)
        net.load_pretrained_model(device=device, model_path=model_path)
        net.eval()

        for module in net.module_set:
            module.load_pretrained_model(device=device, model_path=module_paths[module])
            module.eval()

        net = BaseEmbeddingPytorchClassifier(
            model=net,
            preprocessing=preprocessing,
            postprocessing=postprocessing,
            trainer=trainer,
            threshold=threshold,
        )
        return net
    
    def embed(self, x: torch.Tensor):
        """
        Embedding function in Quo Vadis is forward pass over all early fusion modules.
        """
        embeddings = []
        for module in self.module_set:
            embeddings.append(module.forward(x))
        return torch.cat(embeddings, dim=1)
    
    def embedding_layer(self):
        raise NotImplementedError("QuoVadis does not have a single embedding layer")
    
    def embedding_matrix(self):
        raise NotImplementedError("QuoVadis does not have a single embedding matrix")

    def _forward_embed_x(self, x: torch.Tensor):
        return self.quovadis_meta_model(x)
    
    def forward(self, x: torch.Tensor):
        x = self.embed(x)
        return self._forward_embed_x(x)



import requests


def download_gdrive(gdrive_id, fname_save):
    """Extracted from RobustBench sourcecode https://github.com/RobustBench/robustbench.git."""

    def get_confirm_token(response):
        for key, value in response.cookies.items():
            if key.startswith("download_warning"):
                return value

        return None
    
    def get_uuid_token(response):
        # TODO: rude way, might want to implement proper HTML parsing
        if 'uuid" value="' not in response.text:
            return None
        
        uuid = response.text.split("uuid")[1].replace('" value="', "").split('"')[0]
        return uuid

    def save_response_content(response, fname_save):
        CHUNK_SIZE = 32768

        with open(fname_save, "wb") as f:
            for chunk in response.iter_content(CHUNK_SIZE):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)

    print("Download started: path={} (gdrive_id={})".format(fname_save, gdrive_id))

    docs_domain = "https://docs.google.com/uc"
    drive_domain = "https://drive.usercontent.google.com/download"
    session = requests.Session()

    response = session.get(docs_domain, params={"id": gdrive_id, "export": "download", "confirm": "t"}, stream=True)
    token = get_confirm_token(response)
    uuid = get_uuid_token(response)

    if token:
        params = {"id": gdrive_id, "confirm": token,  "export": "download"}
        response = session.get(docs_domain, params=params, stream=True)
    elif uuid:
        params = {"id": gdrive_id, "uuid": uuid, "export": "download", "confirm": "t"}
        response = session.get(drive_domain, params=params, stream=True)

    save_response_content(response, fname_save)
    session.close()
    print("Download finished: path={} (gdrive_id={})".format(fname_save, gdrive_id))



import math


def calculate_delta(chunk_size: int, manipulations: dict) -> int:
    """
    Calculate the total delta for the certification procedure.

    Parameters
    ----------
    chunk_size : int
        The size of the chunk.
    manipulations : dict
        The manipulations dictionary. It contains the number of bytes patched, appended, and injected.

    Returns
    -------
    int
        The total delta.
    """
    if manipulations["num_bytes_patched"] is not None:
        patch_delta = sum(
            [math.ceil(adv_content / chunk_size) + 1 for adv_content in manipulations["num_bytes_patched"]])
    else:
        patch_delta = 0

    if manipulations["num_bytes_appended"] is not None:
        append_delta = sum(
            [math.ceil(adv_content / chunk_size) + 1 for adv_content in manipulations["num_bytes_appended"]])
    else:
        append_delta = 0

    if manipulations["num_bytes_injected"] is not None:
        inject_delta = sum(
            [math.ceil(adv_content / chunk_size) + 1 for adv_content in manipulations["num_bytes_injected"]])
    else:
        inject_delta = 0

    total_delta = (2 * patch_delta) + append_delta + inject_delta
    return total_delta

def certify_EXEmple(labels: list, chunk_size: int, manipulations: dict) -> bool:
    """
    Certify whether the final classification is robust to patch and injection manipulations based on the following work:
    Daniel Gibert, Luca Demetrio, Giulio Zizzo, Quan Le, Jordi Planes, Battista Biggio
    Certified Adversarial Robustness of Machine Learning-based Malware Detectors via (De) Randomized Smoothing
    ArXiv'24

    Parameters
    ----------
    labels : list
        The list of labels.
    chunk_size : int
        The size of the chunk.
    manipulations : dict
        The manipulations dictionary. It contains the number of bytes patched, appended, and injected.

    Returns
    -------
    bool
        Certification guarantee
    """
    total_delta = calculate_delta(chunk_size, manipulations)
    num_top_class = sum([1 for label in labels if label >= 0.5])
    num_second_class = sum([1 for label in labels if label < 0.5])

    if num_top_class > num_second_class + total_delta:
        is_certified = True
    else:
        is_certified = False
    return is_certified




"""
Trizna, D. and Demetrio, L. and Battista, B. and Fabio, R. (2024)
Nebula: Self-Attention for Dynamic Malware Analysis
IEEE Transactions on Information Forensics and Security (TIFS)
Paper: https://ieeexplore.ieee.org/document/10551436
Source: https://github.com/dtrizna/nebula/blob/main/nebula/preprocessing/tokenization.py
"""

import os
import logging
import json
import numpy as np
from tqdm import tqdm
from time import time
from functools import reduce
from typing import Iterable, Union, List, Dict, Tuple
from collections import Counter, defaultdict
from pandas import json_normalize, concat, DataFrame, merge

from nltk import WhitespaceTokenizer, WordPunctTokenizer
import sentencepiece as spm

from .constants import JSON_CLEANUP_SYMBOLS, SPEAKEASY_TOKEN_STOPWORDS
import string


class JSONFilter:
    def __init__(self, fields: List[str], normalized: bool = False):
        if not isinstance(fields, list) or not all(isinstance(f, str) for f in fields):
            raise TypeError("fields must be a list of strings")
        if not isinstance(normalized, bool):
            raise TypeError("normalized must be a boolean")
        self.fields = fields
        self.normalized = normalized

    @staticmethod
    def filter_non_normalized_field(
        json_event: Union[List[Dict], Dict],
        field: str
    ) -> Tuple[Union[DataFrame, None], List[str]]:
        keys = field.split(".")
        current_val = json_event
        keys_iterated = []

        for key in keys:
            keys_iterated.append(key)
            if isinstance(current_val, dict):
                if key in current_val:
                    current_val = current_val[key]
                else:
                    return None, keys_iterated
            elif isinstance(current_val, list):
                table = json_normalize(json_event, record_path=keys_iterated)[keys[len(keys_iterated):]]
                return table, keys_iterated
            else:
                return current_val, keys_iterated

        if current_val == []:
            return None, keys_iterated
        elif isinstance(current_val, (dict, list)):
            return current_val, keys_iterated
        else:
            return current_val, keys_iterated

    def filter_non_normalized_event(self, json_event: Union[List[Dict], Dict]) -> Dict:
        values = defaultdict(list)
        for field in self.fields:
            filtered_value, key = self.filter_non_normalized_field(json_event, field)
            if filtered_value is not None:
                values['.'.join(key)].append(filtered_value)

        # Merge tables into a single DataFrame
        for key, value_list in values.items():
            if all(isinstance(x, DataFrame) for x in value_list):
                values[key] = reduce(
                    lambda x, y: merge(x, y, left_index=True, right_index=True),
                    value_list
                )
        return values

    def filter_normalized_event(self, json_event: Union[List[Dict], Dict]) -> DataFrame:
        table = json_normalize(json_event)
        # Preserve only the specified fields
        cols = table.columns[table.columns.isin(self.fields)]
        table = table[cols]
        return table

    def filter(self, json_events: Union[List[Dict], Dict]) -> Union[List[DataFrame], List[Dict]]:
        if isinstance(json_events, str):
            json_events = json.loads(json_events)
        if not isinstance(json_events, (list, dict)):
            raise TypeError("json_events must be a list or dict!")
        if isinstance(json_events, dict):
            json_events = [json_events]
        if not all(isinstance(x, dict) for x in json_events):
            raise TypeError("json_events must be a list of dicts!")

        if self.normalized:
            filtered_events = [self.filter_normalized_event(x) for x in json_events]
            return filtered_events  # List of DataFrames
        else:
            filtered_events = [self.filter_non_normalized_event(x) for x in json_events]
            return filtered_events  # List of dicts

    def filter_and_concat(self, json_events: Union[List[Dict], Dict]) -> Dict:
        filtered_events = self.filter(json_events)
        record_dict = defaultdict(DataFrame)
        for table_dict in filtered_events:
            for key in table_dict:
                record_dict[key] = concat([record_dict[key], table_dict[key]], axis=0, ignore_index=True)
        return record_dict


class JSONTokenizer:
    def __init__(
        self,
        seq_len: int,
        cleanup_symbols: List[str] = JSON_CLEANUP_SYMBOLS,
        stopwords: List[str] = SPEAKEASY_TOKEN_STOPWORDS,
        special_tokens: List[str] = None
    ):
        if not isinstance(seq_len, int):
            raise TypeError("seq_len must be an integer!")
        self.seq_len = seq_len

        if cleanup_symbols is not None and not isinstance(cleanup_symbols, (list, tuple)):
            raise TypeError("cleanup_symbols must be a list or tuple!")
        self.cleanup_symbols = cleanup_symbols

        if stopwords is not None and not isinstance(stopwords, (list, tuple)):
            raise TypeError("stopwords must be a list or tuple!")
        self.stopwords = stopwords

        if special_tokens is None:
            special_tokens = ["<pad>", "<unk>", "<mask>"]
        self.special_tokens = dict(zip(special_tokens, range(len(special_tokens))))
        if len(self.special_tokens) < 3:
            raise ValueError("special_tokens must contain at least 3 tokens for pad, unk, and mask!")
        self.pad_token = special_tokens[0]
        self.unk_token = special_tokens[1]
        self.mask_token = special_tokens[2]
        self.pad_token_id = self.special_tokens[self.pad_token]
        self.unk_token_id = self.special_tokens[self.unk_token]
        self.mask_token_id = self.special_tokens[self.mask_token]

        self.vocab = None
        self.reverse_vocab = None

    def clear_json_event(self, text: Union[str, bytes, list, dict]) -> str:
        """
        Removes all special characters from the JSON event.
        """
        if not isinstance(text, (str, bytes, list, dict)):
            raise TypeError("Input must be a string, bytes, list, or dict!")
        if isinstance(text, bytes):
            text = text.decode('utf-8')
        elif isinstance(text, list) and all(isinstance(item, str) for item in text):
            text = ' '.join(text)
        else:
            text = str(text)
        text = text.lower()
        if self.cleanup_symbols:
            for pattern in self.cleanup_symbols:
                text = text.replace(pattern, ' ')
        if self.stopwords:
            for pattern in self.stopwords:
                text = text.replace(pattern, '')
        # Replace all other punctuation with space
        text = text.translate(str.maketrans(string.punctuation, ' ' * len(string.punctuation)))
        return text

    def pad_sequence(self, encoded_sequence: np.ndarray, seq_len: int = None):
        seq_len = seq_len if seq_len is not None else self.seq_len
        if not isinstance(seq_len, int):
            raise TypeError("seq_len must be an integer!")
        if len(encoded_sequence) >= seq_len:
            return encoded_sequence[:seq_len]
        else:
            padded = np.pad(
                encoded_sequence,
                (0, seq_len - len(encoded_sequence)),
                mode='constant',
                constant_values=self.pad_token_id
            )
            return padded

    def pad_sequence_list(self, encoded_sequence_list: List[np.ndarray], seq_len=None):
        return np.array([self.pad_sequence(x, seq_len) for x in encoded_sequence_list], dtype=np.int32)

    def pad_sequences(self, encoded_sequences: List[np.ndarray], seq_len=None):
        return self.pad_sequence_list(encoded_sequences, seq_len=seq_len)


class JSONTokenizerNaive(JSONTokenizer):
    def __init__(
        self,
        seq_len: int,
        vocab_size: int = 10000,
        vocab: Union[Dict, str] = None,
        cleanup_symbols: List[str] = JSON_CLEANUP_SYMBOLS,
        stopwords: List[str] = SPEAKEASY_TOKEN_STOPWORDS,
        tokenizer_type: str = "whitespace",
        counter_dump: bool = False
    ):
        super().__init__(
            seq_len,
            cleanup_symbols,
            stopwords
        )
        self.vocab_size = vocab_size
        if tokenizer_type not in ["whitespace", "wordpunct"]:
            raise ValueError("tokenizer_type must be either 'whitespace' or 'wordpunct'!")
        if tokenizer_type == "whitespace":
            self.tokenizer = WhitespaceTokenizer()
        elif tokenizer_type == "wordpunct":
            self.tokenizer = WordPunctTokenizer()
        self.counter = None
        self.counter_dump = counter_dump
        self.vocab_error = "Vocabulary not initialized! Use build_vocab() first or load it using load_vocab()!"
        if vocab is not None:
            if vocab == "quovadis":
                vocab = os.path.join(os.path.dirname(__file__), "quovadis_apis.json")
            self.load_vocab(vocab)

    def tokenize_event(self, json_event):
        json_event_clean = self.clear_json_event(json_event)
        tokenized_json_event = self.tokenizer.tokenize(json_event_clean)
        return tokenized_json_event

    def tokenize(self, sample):
        if isinstance(sample, dict):
            return self.tokenize_event(str(sample))
        elif isinstance(sample, (str, bytes)):
            return self.tokenize_event(sample)
        elif isinstance(sample, Iterable):
            return [self.tokenize_event(str(x)) for x in sample]
        else:
            raise TypeError("tokenize(): Input must be a string, bytes, or Iterable!")

    def build_vocab(self, corpus, vocab_size=None, model_prefix="whitespace", counter_dump=False):
        """Builds the vocabulary from the corpus and preserves the
        top vocab_size tokens based on appearance counts."""
        if vocab_size:
            self.vocab_size = vocab_size

        self.counter = Counter()
        for text in tqdm(corpus):
            text = self.clear_json_event(text)
            tokens = self.tokenizer.tokenize(text)
            self.counter.update(tokens)

        # Preserve the most common tokens
        most_common_tokens = self.counter.most_common(self.vocab_size - len(self.special_tokens))
        vocab_tokens = [token for token, _ in most_common_tokens]
        self.vocab = {token: index for index, token in enumerate(self.special_tokens.keys() + vocab_tokens)}
        self.reverse_vocab = {index: token for token, index in self.vocab.items()}
        self.dump_vocab(model_prefix)
        if counter_dump or self.counter_dump:
            self.dump_counter(model_prefix)

    def train(self, corpus, vocab_size=None, model_prefix="whitespace", counter_dump=False):
        self.build_vocab(corpus, vocab_size, model_prefix, counter_dump)

    def dump_vocab(self, vocab_prefix="whitespace"):
        with open(f"{vocab_prefix}_vocab.json", "w") as f:
            json.dump(self.vocab, f, indent=4)
        logging.info(f"Dumped vocab to {vocab_prefix}_vocab.json")

    def dump_counter(self, prefix):
        file = f"{prefix}_counter.json"
        with open(file, "w") as f:
            json.dump(self.counter, f, indent=4)
        logging.info(f"Dumped vocab counter to {file}")

    def load_vocab(self, vocab):
        if isinstance(vocab, dict):
            self.vocab = vocab
        else:
            with open(vocab) as f:
                self.vocab = json.load(f)
        self.vocab_size = len(self.vocab)
        self.reverse_vocab = {v: k for k, v in self.vocab.items()}

    def load_from_pretrained(self, vocab):
        self.load_vocab(vocab)

    def convert_tokens_to_ids(self, tokens):
        if self.vocab:
            return [self.vocab.get(token, self.vocab.get("<unk>")) for token in tokens]
        else:
            raise Exception("convert_tokens_to_ids(): " + self.vocab_error)

    def encode(self, inputs, pad=True, tokenize=True):
        if tokenize:
            inputs = self.tokenize(inputs)
        if isinstance(inputs[0], str):
            # Single sequence
            encoded = [self.convert_tokens_to_ids(inputs)]
        else:
            encoded = [self.convert_tokens_to_ids(seq) for seq in inputs]
        if pad:
            return self.pad_sequence_list(encoded)
        else:
            return encoded

    def decode(self, encoded_sequence):
        if self.vocab and self.reverse_vocab:
            decoded_sequence = []
            for x in encoded_sequence:
                if x == self.pad_token_id:
                    break
                decoded_sequence.append(self.reverse_vocab.get(x, self.unk_token))
            return decoded_sequence
        else:
            raise Exception("decode(): " + self.vocab_error)


class JSONTokenizerBPE(JSONTokenizer):
    def __init__(
        self,
        vocab_size,
        seq_len,
        model_path=None,
        vocab=None,
        cleanup_symbols=JSON_CLEANUP_SYMBOLS,
        stopwords=SPEAKEASY_TOKEN_STOPWORDS,
    ):
        super().__init__(
            vocab_size,
            seq_len,
            cleanup_symbols,
            stopwords
        )
        self.model_path = model_path
        if model_path is not None:
            self.tokenizer = spm.SentencePieceProcessor(model_file=f"{model_path}.model")
            logging.info("Successfully loaded pre-trained tokenizer model!")
            self.load_vocab(vocab=vocab)
        else:
            self.tokenizer = spm.SentencePieceTrainer
            logging.warning(
                "Initialized tokenizer without pre-trained model. "
                "You need to train tokenizer with .train() or specify 'model_path=' during initialization!"
            )

    @staticmethod
    def split_string_to_chunks(s: str, chunk_size: int = 4192):
        """Splits a long string into smaller chunks of size chunk_size, avoiding splitting in the middle of a word."""
        words = s.split(" ")
        chunks = []
        current_chunk = ""
        for word in words:
            if len(current_chunk) + len(word) < chunk_size:
                current_chunk += word + " "
            else:
                chunks.append(current_chunk.strip())
                current_chunk = word + " "
        chunks.append(current_chunk.strip())
        return chunks

    def load_vocab(self, vocab: Union[dict, str, None] = None) -> None:
        if isinstance(vocab, dict):
            self.vocab = vocab
            self.reverse_vocab = {v: k for k, v in self.vocab.items()}
            return

        if vocab is None:
            vocab = f"{self.model_path}_vocab.json"
        if not os.path.exists(vocab):
            vocab = f"{self.model_path}.vocab"
        if not os.path.exists(vocab):
            logging.error(f"Vocab file {vocab} does not exist! .load_vocab() failed!")
            return

        with open(vocab, encoding="utf-8") as f:
            if vocab.endswith(".json"):
                self.vocab = json.load(f)
            else:
                data = f.read()
                vocab_list = [x.split("\t")[0] for x in data.strip().split("\n")]
                self.vocab = {k: i for i, k in enumerate(vocab_list)}
        # Update vocab with special tokens
        for k, v in self.special_tokens.items():
            self.vocab[k] = v
        self.reverse_vocab = {v: k for k, v in self.vocab.items()}
        logging.info(f"Loaded vocab from {vocab}")

    def dump_vocab(self):
        vocab_file_name = f"{self.model_path}_vocab.json"
        with open(vocab_file_name, "w") as f:
            json.dump(self.vocab, f, indent=4)
        logging.info(f"Dumped vocab to {vocab_file_name}")

    def train(
        self,
        json_data,
        vocab_size=None,
        model_prefix="bpe",
        model_type="bpe",
        split_by_number=False,
        sp_length=4192,
        remove_train_files=True
    ):
        """
        Trains the tokenizer on the given JSON data.
        """
        logging.info("Preparing data for SentencePiece tokenizer...")
        json_data_clean = self.clear_json_event(json_data)
        # Split the string into chunks due to SentencePiece limitations
        json_data_chunks = self.split_string_to_chunks(json_data_clean.replace("\\\\", "\\"), chunk_size=sp_length)
        train_file = f"{model_prefix}_trainset_{int(time())}.txt"
        with open(train_file, "w", encoding="utf-8") as f:
            f.write("\n".join(json_data_chunks))

        if vocab_size:
            self.vocab_size = vocab_size

        train_cmd = " ".join([
            f"--input={train_file}",
            f"--model_prefix={model_prefix}",
            f"--vocab_size={self.vocab_size}",
            f"--model_type={model_type}",
            f"--split_by_number={split_by_number}",
            f"--max_sentence_length={sp_length}",
            "--max_sentencepiece_length=64"
        ])
        logging.info(f"Training tokenizer with command: {train_cmd}")
        self.tokenizer.Train(train_cmd)
        self.tokenizer = spm.SentencePieceProcessor(model_file=f"{model_prefix}.model")

        self.model_path = model_prefix
        self.load_vocab()
        self.dump_vocab()

        if remove_train_files:
            os.remove(train_file)
            os.remove(f"{model_prefix}.vocab")

    def tokenize(self, inputs):
        """
        Tokenizes the given JSON data.
        """
        if isinstance(inputs, (str, bytes, dict)):
            inputs = [inputs]
        data_clean = [self.clear_json_event(x) for x in inputs]
        return [self.tokenizer.encode_as_pieces(x) for x in data_clean]

    def encode(self, inputs, pad=True, tokenize=True):
        if not tokenize:
            raise NotImplementedError("SentencePiece tokenizer does not support encode without tokenize!")
        if isinstance(inputs, (str, bytes, dict)):
            inputs = [inputs]

        data_clean = [self.clear_json_event(x) for x in inputs]
        encoded = [self.tokenizer.encode_as_ids(x) for x in data_clean]
        if pad:
            return self.pad_sequence_list(encoded)
        else:
            return encoded



import copy
from typing import Optional

import torch
from torch import Tensor


class DifferentiableOneHotEncoding(torch.nn.Module):
    def __init__(self, classes):
        super(DifferentiableOneHotEncoding, self).__init__()
        self.num_classes = classes

    def forward(self, x):
        eye = torch.eye(self.num_classes, device=x.device)
        one_hot = torch.index_select(eye, dim=0, index=x.view(-1))
        one_hot = one_hot.view(x.shape[0], x.shape[1], self.num_classes)
        return one_hot


class IdentityEmbedding(torch.nn.Embedding):
    def __init__(
        self,
        num_embeddings: int,
        embedding_dim: int,
        padding_idx: Optional[int] = None,
        max_norm: Optional[float] = None,
        norm_type: float = 2.0,
        scale_grad_by_freq: bool = False,
        sparse: bool = False,
        _weight: Optional[Tensor] = None,
        _freeze: bool = False,
        device=None,
        dtype=None,
    ):
        super().__init__(
            num_embeddings,
            embedding_dim,
            padding_idx,
            max_norm,
            norm_type,
            scale_grad_by_freq,
            sparse,
            _weight,
            _freeze,
            device,
            dtype,
        )

        self.encoder = DifferentiableOneHotEncoding(classes=self.num_embeddings)

    def forward(self, x: Tensor) -> Tensor:
        x_encoded = self.encoder(x)
        emb = x_encoded.matmul(self.weight)
        return emb


def patch_embedding_backward(model: torch.nn.Module):
    patched_model = copy.deepcopy(model)
    modules = patched_model.named_modules()
    for name, module in modules:
        if isinstance(module, torch.nn.Embedding):
            new_embedding_layer = IdentityEmbedding(
                num_embeddings=module.num_embeddings,
                embedding_dim=module.embedding_dim,
                padding_idx=module.padding_idx,
                max_norm=module.max_norm,
                norm_type=module.norm_type,
                scale_grad_by_freq=module.scale_grad_by_freq,
                sparse=module.sparse,
                device=module.weight.device,
                _weight=module.weight,
            )
            setattr(patched_model, name, new_embedding_layer)
    return patched_model



import torch

from maltorch.initializers.initializers import ByteBasedInitializer


class PartialDOSInitializer(ByteBasedInitializer):
    def __init__(self, random_init: bool = False):
        super().__init__(random_init)

    def __call__(self, x: torch.Tensor):
        indexes = torch.arange(2, 60).long().repeat((x.shape[0], 1))
        delta = (
            torch.zeros((x.shape[0], 58))
            if not self.random_init
            else torch.randint(0, 255, (x.shape[0], 58))
        ).float()
        return x, delta, indexes



import importlib
from abc import abstractmethod
from typing import Type

from maltorch.adv.evasion.backend_attack import BackendAttack


class OptimizerBackends:
    """Available backends."""

    NG = "nevergrad"
    GRADIENT = "gradient"


class BaseOptimAttackCreator:
    """Generic creator for attacks."""

    @classmethod
    def get_implementation(cls, backend: str) -> Type[BackendAttack]:
        """
        Get the implementation of the attack with the given backend.

        Parameters
        ----------
        backend : str
            The backend for the attack. See secmlt.adv.backends for
            available backends.

        Returns
        -------
        BaseEvasionAttack
            Attack implementation.
        """
        implementations = {
            OptimizerBackends.NG: cls.get_nevergrad_implementation,
            OptimizerBackends.GRADIENT: cls._get_native_implementation,
        }
        cls.check_backend_available(backend)
        return implementations[backend]()

    @classmethod
    def check_backend_available(cls, backend: str) -> bool:
        """
        Check if a given backend is available for the attack.

        Parameters
        ----------
        backend : str
            Backend string.

        Returns
        -------
        bool
            True if the given backend is implemented.

        Raises
        ------
        NotImplementedError
            Raises NotImplementedError if the requested backend is not in
            the list of the possible backends (check secmlt.adv.backends).
        """
        if backend in cls.get_backends():
            return True
        msg = "Unsupported or not-implemented backend."
        raise NotImplementedError(msg)

    @classmethod
    def get_nevergrad_implementation(cls) -> BackendAttack:
        """
        Get the Nevergrad implementation of the attack.

        Returns
        -------
        BaseEvasionAttack
            Nevergrad implementation of the attack.

        Raises
        ------
        ImportError
            Raises ImportError if Nevergrad extra is not installed.
        """
        if importlib.util.find_spec("nevergrad", None) is not None:
            return cls._get_nevergrad_implementation()
        msg = "Nevergrad extra not installed."
        raise ImportError(msg)

    @staticmethod
    def _get_nevergrad_implementation() -> BackendAttack:
        msg = "Nevergrad implementation not available."
        raise NotImplementedError(msg)

    @staticmethod
    def _get_native_implementation() -> BackendAttack:
        msg = "Native implementation not available."
        raise NotImplementedError(msg)

    @staticmethod
    @abstractmethod
    def get_backends() -> set[str]:
        """
        Get the available backends for the given attack.

        Returns
        -------
        set[str]
            Set of implemented backends available for the attack.

        Raises
        ------
        NotImplementedError
            Raises NotImplementedError if not implemented in the inherited class.
        """
        msg = "Backends should be specified in inherited class."
        raise NotImplementedError(msg)



import abc

import torch
from secmlt.manipulations.manipulation import Manipulation
from secmlt.optimization.constraints import Constraint

from maltorch.initializers.initializers import ByteBasedInitializer


class ByteManipulation(Manipulation, abc.ABC):
    def __init__(
        self,
        initializer: ByteBasedInitializer,
        domain_constraints: list[Constraint],
        perturbation_constraints: list[Constraint],
    ):
        super().__init__(domain_constraints, perturbation_constraints)
        self.initializer = initializer
        self.indexes_to_perturb = []

    @abc.abstractmethod
    def initialize(self, samples: torch.Tensor) -> [torch.Tensor, torch.Tensor]: ...



import torch
from secmlt.models.data_processing.data_processing import DataProcessing


class RandomizedDeletionPreprocessing(DataProcessing):
    """
    Zhougun Huang, Niel Marchant, Keane Lucas, Lujo Bauer, Olya Ohrimenko, Benjamin I. P. Rubenstein
    RS-Del: Edit Distance Robustness Certificates for Sequence Classifiers via Randomized Deletion
    NeurIPS 2023,
    """
    def __init__(self, pdel: float = 0.03, num_versions: int = 100, padding_idx: int = 256):
        super().__init__()
        self.pdel = pdel
        self.num_versions = num_versions
        self.padding_idx = padding_idx

    def _process(self, x: torch.Tensor) -> torch.Tensor:
        x = x.squeeze()  # Remove all dimensions equal to 1
        vecs = []
        for i in range(self.num_versions):
            mask = torch.rand(x.shape[0]) > self.pdel
            mask = mask.to(x.device)
            masked_x = torch.masked_select(x, mask=mask)
            vecs.append(masked_x)
        x = torch.nn.utils.rnn.pad_sequence(vecs, batch_first=True, padding_value=self.padding_idx)
        return x

    def invert(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the inverted transform (if defined).

        Parameters
        ----------
        x : torch.Tensor
            Input samples.

        Returns
        -------
        torch.Tensor
            The samples in the input space before the transformation.
        """
        raise NotImplementedError




from typing import Union, List, Callable

import torch
from secmlt.adv.evasion.base_evasion_attack import BaseEvasionAttack
from secmlt.models.base_model import BaseModel
from secmlt.optimization.initializer import Initializer
from secmlt.trackers import Tracker
from torch.utils.data import TensorDataset, DataLoader

from maltorch.manipulations.byte_manipulation import ByteManipulation


class BackendAttack(BaseEvasionAttack):
    @classmethod
    def _trackers_allowed(cls):
        return True

    @staticmethod
    def get_perturbation_models():
        pass

    def __init__(
        self,
        y_target: Union[int, None],
        query_budget: int,
        loss_function: Union[str, torch.nn.Module],
        optimizer_cls: Callable,
        manipulation_function: ByteManipulation,
        initializer: Initializer,
        trackers: Union[List[Tracker], Tracker] = None,
        **kwargs
    ):
        self.y_target = y_target
        self.query_budget = query_budget
        self.loss_function = loss_function
        self.manipulation_function = manipulation_function
        self.initializer = initializer
        self.trackers = trackers
        self.optimizer = None
        self.optimizer_cls = optimizer_cls
        self._best_loss = None
        self._best_delta = None

    def _init_attack_manipulation(
        self, samples: torch.Tensor
    ) -> (torch.Tensor, torch.Tensor):
        return self.manipulation_function.initialize(samples.data)

    def _apply_manipulation(
        self, x: torch.Tensor, delta: torch.Tensor
    ) -> (torch.Tensor, torch.Tensor):
        raise NotImplementedError()

    def _optimizer_step(self, delta: torch.Tensor, loss: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()

    def _init_optimizer(self, model: BaseModel, delta: torch.Tensor) -> Callable:
        raise NotImplementedError()

    def _consumed_budget(self):
        raise NotImplementedError()

    def _init_best_tracking(self, delta: torch.Tensor):
        self._best_delta = torch.zeros_like(delta.detach().cpu())
        self._best_loss = torch.zeros((delta.shape[0], 1)).fill_(torch.inf)

    def _track_best(self, loss: torch.Tensor, delta: torch.Tensor):
        delta_to_track = delta.cpu()
        where_best = (loss.cpu() < self._best_loss.cpu()).squeeze(1)
        self._best_delta[where_best] = delta_to_track[where_best]
        self._best_loss[where_best] = loss.cpu()[where_best]

    def _get_best_delta(self):
        return self._best_delta

    def _track(
        self,
        iteration: int,
        loss: torch.Tensor,
        scores: torch.Tensor,
        x_adv: torch.Tensor,
        delta: torch.Tensor,
    ):
        if self._trackers_allowed():
            if self.trackers:
                for tracker in self.trackers:
                    tracker.track(
                        iteration=iteration,
                        loss=loss.data,
                        scores=scores.data,
                        x_adv=x_adv.data,
                        delta=delta,
                        grad=None,
                    )

    def __call__(self, model: BaseModel, data_loader: DataLoader) -> DataLoader:
        """
        Compute the attack against the model, using the input data.

        Parameters
        ----------
        model : BaseModel
            Model to test.
        data_loader : DataLoader
            Test dataloader.

        Returns
        -------
        DataLoader
            Dataloader with adversarial examples and original labels.
        """
        adversarials = []
        original_labels = []
        for samples, labels in data_loader:
            x_adv, _ = self._run(model, samples, labels)
            adversarials.append(x_adv)
            original_labels.append(labels)
        adversarials = torch.vstack(adversarials)
        original_labels = torch.vstack(original_labels)
        adversarial_dataset = TensorDataset(adversarials, original_labels)
        return DataLoader(
            adversarial_dataset,
            batch_size=data_loader.batch_size,
        )

    def _run(
        self,
        model: BaseModel,
        samples: torch.Tensor,
        labels: torch.Tensor,
        **optim_kwargs,
    ) -> (torch.Tensor, torch.Tensor):
        multiplier = 1 if self.y_target is not None else -1
        target = (
            torch.zeros_like(labels) + self.y_target
            if self.y_target is not None
            else labels
        ).type(labels.dtype)
        target = target.to(labels.device)
        x_adv, delta = self._init_attack_manipulation(samples)
        self.optimizer = self._init_optimizer(model, delta)
        budget = 0
        self._init_best_tracking(delta)
        while budget < self.query_budget:
            x_adv, _ = self._apply_manipulation(samples, delta)
            scores = model.decision_function(x_adv)
            loss = self.loss_function(scores, target) * multiplier
            delta = self._optimizer_step(delta, loss)
            budget += self._consumed_budget()
            self._track(budget, loss, scores, x_adv, delta)
            self._track_best(loss, delta)
        best_delta = self._get_best_delta()
        best_x, _ = self._apply_manipulation(samples, best_delta)
        return best_x, self._best_delta




from torchvision.models import resnet18

from maltorch.zoo.model import PytorchModel


class ResNet18(PytorchModel):
    def __init__(self, threshold: float = 0.5):
        super(ResNet18, self).__init__(
            name="ResNet18", gdrive_id="1N1uK8bsfJvB88ryZcbRfxkzerblXqMqg"
        )
        self.model = resnet18(num_classes=1)
        self.threshold = threshold

    def forward(self, x):
        y = self.model(x)
        return y



"""
Malware Detection by Eating a Whole EXE
Edward Raff, Jon Barker, Jared Sylvester, Robert Brandon, Bryan Catanzaro, Charles Nicholas
https://arxiv.org/abs/1710.09435
"""

import torch
import torch.nn.functional as F
from torch import nn

from maltorch.zoo.model import EmbeddingModel


class MalConv(EmbeddingModel):
    """
    Architecture implementation.
    """

    def __init__(self, embedding_size: int = 8,
                 max_len: int =2**20,
                 threshold: float =0.5,
                 padding_idx: int = 256,
                 kernel_size=500,
                 stride=500):
        super(MalConv, self).__init__(
            name="MalConv", gdrive_id="1Hg8I7Jx13LmnSPBjsPGr8bvmmS874Y9N"
        )
        self.embedding_1 = nn.Embedding(
            num_embeddings=257, embedding_dim=embedding_size, padding_idx=padding_idx
        )
        self.conv1d_1 = nn.Conv1d(
            in_channels=embedding_size,
            out_channels=128,
            kernel_size=(kernel_size,),
            stride=(stride,),
            groups=1,
            bias=True,
        )
        self.conv1d_2 = nn.Conv1d(
            in_channels=embedding_size,
            out_channels=128,
            kernel_size=(kernel_size,),
            stride=(stride,),
            groups=1,
            bias=True,
        )
        self.dense_1 = nn.Linear(in_features=128, out_features=128, bias=True)
        self.dense_2 = nn.Linear(in_features=128, out_features=1, bias=True)
        self.embedding_size = (embedding_size,)
        self.max_len = max_len
        self.threshold = threshold
        self.invalid_value = padding_idx

    def embedding_layer(self):
        return self.embedding_1

    def embed(self, x):
        emb_x = self.embedding_1(x)
        emb_x = emb_x.transpose(1, 2)
        return emb_x

    def _forward_embed_x(self, x):
        conv1d_1 = self.conv1d_1(x)
        conv1d_2 = self.conv1d_2(x)
        conv1d_1_activation = torch.relu(conv1d_1)
        conv1d_2_activation = torch.sigmoid(conv1d_2)
        multiply_1 = conv1d_1_activation * conv1d_2_activation
        global_max_pooling1d_1 = F.max_pool1d(
            input=multiply_1, kernel_size=multiply_1.size()[2:]
        )
        global_max_pooling1d_1_flatten = global_max_pooling1d_1.view(
            global_max_pooling1d_1.size(0), -1
        )
        dense_1 = self.dense_1(global_max_pooling1d_1_flatten)
        dense_1_activation = torch.relu(dense_1)
        dense_2 = self.dense_2(dense_1_activation)
        return dense_2

    def embedding_matrix(self):
        return self.embedding_1.weight



import torch
from maltorch.datasets.binary_dataset import BinaryDataset


class RandomizedAblationDataset(BinaryDataset):
    """
    Daniel Gibert, Giulio Zizzo, Quan Le
    Towards a Practical Defense Against Adversarial Attacks on Deep Learning-Based Malware Detectors via Randomized
    Smoothing.
    ESORICS Workshops, SECAI 2024
    """
    def __init__(self,
                 csv_filepath: str = None,
                 goodware_directory: str = None,
                 malware_directory: str = None,
                 max_len: int = 2**20,
                 padding_idx: int = 256,
                 min_len: int = None,
                 num_versions: int = 100,
                 pabl: float = 0.20,
                 is_training: bool = True):
        super().__init__(
            csv_filepath=csv_filepath,
            goodware_directory=goodware_directory,
            malware_directory=malware_directory,
            max_len=max_len,
            padding_idx=padding_idx,
            min_len=min_len
        )
        self.num_versions = num_versions
        self.pabl = pabl
        self.is_training = is_training

    def pad_collate_func(self, batch):
        """
        This function randomly ablates the bytes given a probability pabl.
        It works differently at training and at test time.
        During training, given an input example we generate an ablated version of that example.
        During testing, given an input example we generate N ablated versions of that example.
        """
        vecs = []
        labels = []
        if self.is_training is True:
            for x, y in batch:
                # Get mask
                mask = torch.rand(x.shape[0]) <= self.pabl
                # Apply mask - Convert masked elements to self.padding_idx
                masked_x = x.masked_fill(mask, self.padding_idx)
                vecs.append(masked_x)
                labels.append(y)
            x = torch.nn.utils.rnn.pad_sequence(vecs, batch_first=True, padding_value=self.padding_idx)
            # stack will give us (B, 1), so index [:,0] to get to just (B)
            y = torch.tensor(labels)
            return x, y
        else: # Set self.is_training to False whenever you want to generate N versions for each executable.
            if len(batch) == 1:  # Only implemented for batch sizes equals to 1
                x = batch[0][0]
                for i in range(self.num_versions):
                    # Get mask
                    mask = torch.rand(x.shape[0]) <= self.pabl
                    # Apply mask - Convert masked elements to self.padding_idx
                    masked_x = x.masked_fill(mask, self.padding_idx)
                    vecs.append(masked_x)
                x = torch.nn.utils.rnn.pad_sequence(vecs, batch_first=True, padding_value=self.padding_idx)
                y = batch[0][1]
                return x, y
            else:
                raise NotImplementedError




from pathlib import Path

import torch


def load_from_folder(
    path: Path, extension: str = "exe", padding: int = 256, limit=None
) -> torch.Tensor:
    """Create a torch.Tensor whose rows are all the file with extension specified in input.
    Tensor are padded to match the same size.
    :param path: Folder path
    :param extension: default "exe", filters all the file based on this extension
    :param padding: default 256, pad every tensor with this value to uniform the size
    :param limit: default None, limit the number of loaded file, None for load all folder
    :return: a torch.Tensor containing all the file converted into tensors
    """
    X = []
    for filepath in path.glob(f"*.{extension}"):
        x = load_single_exe(filepath)
        X.append(x)
        if limit is not None and len(X) >= limit:
            break
    X = torch.nn.utils.rnn.pad_sequence(X, padding_value=padding).transpose(0, 1).long()
    return X


def create_labels(x: torch.Tensor, label: int):
    """
    Create the labels for the specified data.
    """
    y = torch.zeros((x.shape[0], 1)) + label
    return y


def load_single_exe(path: Path) -> torch.Tensor:
    """
    Create a torch.Tensor from the file pointed in the path
    :param path: a pathlib Path
    :return: torch.Tensor containing the bytes of the file as a tensor
    """
    with open(path, "rb") as h:
        code = h.read()
    x = torch.frombuffer(bytearray(code), dtype=torch.uint8).to(torch.float)
    return x



import random
import torch
from maltorch.datasets.binary_dataset import BinaryDataset
from abc import ABC, abstractmethod
import torch.nn.functional as F


class DynamicChunkSizeDRSDataset(BinaryDataset, ABC):
    """
    Daniel Gibert, Giulio Zizzo, Quan Le, Jordi Planes
    Adversarial Robustness of Deep Learning-based Malware Detectors via (De) Randomized Smoothing
    IEEE Access 2024
    """
    def __init__(self,
                 csv_filepath: str = None,
                 goodware_directory: str = None,
                 malware_directory: str = None,
                 max_len: int = 2 ** 20,
                 padding_idx: int = 256,
                 min_len: int = None,
                 sort_by_size: bool = False,
                 file_percentage: float = 0.05,
                 num_chunks: int = 100,
                 is_training: bool = True,
                 min_chunk_size=500):
        super().__init__(
            csv_filepath=csv_filepath,
            goodware_directory=goodware_directory,
            malware_directory=malware_directory,
            max_len=max_len,
            padding_idx=padding_idx,
            min_len=min_len
        )
        self.is_training = is_training
        self.file_percentage = file_percentage
        self.num_chunks = num_chunks
        self.min_chunk_size = min_chunk_size
        self.sort_by_size = sort_by_size

        if sort_by_size: #  Reorder files by size
            #sorted(self.all_files, key=lambda x: x[2])
            self.all_files.sort(key=lambda filename: filename[2])

    def pad_collate_func(self, batch):
        """
        This function splits a tensor of bytes into chunks.
        It works differently at training and at test time.
        During training, given an input example we randomly select a chunk of bytes
        During testing, given an input example we sequentially extract N chunks of bytes. This chunks may overlap with
        one another depending on the amount of chunks you want to extract and the size of each chunk.
        """
        if self.is_training is True:  # Select a random chunk from a given executable
            return self.generate_training_examples(batch)
        else:  # Split an executable into chunks
            return self.generate_testing_examples(batch)

    def generate_training_examples(self, batch):
        vecs = []
        labels = []
        max_size = max([x.size()[0] for x, y in batch])
        chunk_size = int(-(-(max_size * self.file_percentage) // 1))

        for x, y in batch:
            # Code here
            if x.shape[0] <= chunk_size:
                vecs.append(x)
            else:
                start_location = random.randint(0, max(0, x.shape[0] - chunk_size))
                end_location = start_location + chunk_size
                vecs.append(x[start_location:end_location])
            labels.append(y)

        # Pad each chunk to `min_len` if necessary
        padded_Xs = []
        for x in vecs:
            if x.shape[0] < self.min_len:
                padding_size = self.min_len - x.shape[0]
                # Pad the vector with padding_value (or any other padding value you prefer)
                padded_x = F.pad(x, (0, padding_size), value=self.padding_idx)
                padded_Xs.append(padded_x)
            else:
                padded_Xs.append(x)

        x = torch.nn.utils.rnn.pad_sequence(padded_Xs, batch_first=True, padding_value=self.padding_idx)
        # stack will give us (B, 1), so index [:,0] to get to just (B)
        y = torch.tensor(labels)
        return x, y

    @abstractmethod
    def generate_testing_examples(self, batch):
        ...




from abc import ABC
from typing import Union, Iterable, Dict, Any

from secmlt.optimization.gradient_processing import GradientProcessing
from torch import Tensor
import torch.optim

from maltorch.zoo.model import BaseEmbeddingPytorchClassifier


class BaseByteOptimizer(torch.optim.Optimizer, ABC):
    def __init__(
        self,
        params: Union[Iterable[Tensor], Iterable[Dict[str, Any]]],
        model: BaseEmbeddingPytorchClassifier,
        indexes_to_perturb: torch.LongTensor,
        gradient_processing: GradientProcessing,
        lr: int = 16,
        device: str = "cpu",
    ):
        defaults = {
            "byte_step_size": lr,
            "embedding_matrix": model.embedding_matrix(),
        }
        super().__init__(params, defaults)
        self.gradient_processing = gradient_processing
        self.step_size = lr
        self.embedding_matrix = model.embedding_matrix()
        self.indexes_to_perturb = indexes_to_perturb
        self.device = device
        model.embedding_layer().register_full_backward_hook(self._backward_hook)
        self._embedding_grad = None

    def zero_grad(self, set_to_none: bool = ...) -> None:
        super().zero_grad(set_to_none=set_to_none)
        self._embedding_grad = None

    def _backward_hook(self, module, input_grad, output_grad):
        self._embedding_grad = output_grad



import torch
import random


class RandomChunkSampler(torch.utils.data.sampler.Sampler):
    """
    Samples random "chunks" of a dataset, so that items within a chunk are always loaded together. Useful to keep chunks in similar size groups to reduce runtime.
    """

    def __init__(self, data_source, batch_size):
        """
        data_source: the souce pytorch dataset object
        batch_size: the size of the chunks to keep together. Should generally be set to the desired batch size during training to minimize runtime.
        """
        self.data_source = data_source
        self.batch_size = batch_size

    def __iter__(self):
        n = len(self.data_source)

        data = [x for x in range(n)]

        # Create blocks
        blocks = [data[i:i + self.batch_size] for i in range(0, len(data), self.batch_size)]
        # shuffle the blocks
        random.shuffle(blocks)
        # concatenate the shuffled blocks
        data[:] = [b for bs in blocks for b in bs]

        return iter(data)

    def __len__(self):
        return len(self.data_source)


import itertools
import math
import struct

import lief
import torch


def _shift_pointer_to_section_content(
        liefpe: lief.PE.Binary,
        raw_code: bytearray,
        entry_index: int,
        amount: int,
        pe_shifted_by: int = 0,
) -> bytearray:
    """
    Shifts the section content pointer.

    Parameters
    ----------
    liefpe : lief.PE.Binary
            the binary wrapper by lief
    raw_code : bytearray
            the code of the executable to eprturb
    entry_index : int
            the entry of the section to manipulate
    amount : int
            the shift amount
    pe_shifted_by : int, optional, default 0
            if the PE header was shifted, this value should be set to that amount
    Returns
    -------
    bytearray
            the modified code
    """
    pe_position = liefpe.dos_header.addressof_new_exeheader + pe_shifted_by
    optional_header_size = liefpe.header.sizeof_optional_header
    coff_header_size = 24
    section_entry_length = 40
    size_of_raw_data_pointer = 20
    shift_position = (
            pe_position
            + coff_header_size
            + optional_header_size
            + (entry_index * section_entry_length)
            + size_of_raw_data_pointer
    )
    old_value = struct.unpack("<I", raw_code[shift_position: shift_position + 4])[0]
    new_value = old_value + amount
    new_value = struct.pack("<I", new_value)
    raw_code[shift_position: shift_position + 4] = new_value

    return raw_code


def _shift_pe_header(
        liefpe: lief.PE.Binary, raw_code: bytearray, amount: int
) -> bytearray:
    """
    Shifts the PE header, injecting a default pattern

    Parameters
    ----------
    liefpe : lief.PE.Binary
            the binary wrapper by lief
    raw_code : bytearray
            the code of the executable to perturb
    amount : int
            how much to inject, already aligned to file_alignment

    Returns
    -------
    bytearray
            the modified code
    """
    if amount == 0:
        return raw_code
    pe_position = liefpe.dos_header.addressof_new_exeheader
    # section_alignment = liefpe.optional_header.section_alignment
    # file_alignment = liefpe.optional_header.file_alignment
    raw_code[0x3C:0x40] = struct.pack("<I", pe_position + amount)

    raw_code[pe_position + 60 + 20 + 4: pe_position + 60 + 20 + 4 + 4] = struct.pack(
        "<I", liefpe.optional_header.sizeof_headers + amount
    )
    pattern = itertools.cycle("I love ToucanStrike <3")
    [raw_code.insert(pe_position, ord(next(pattern))) for _ in range(amount)]

    return raw_code


def extend_manipulation(
        x: torch.Tensor, preferable_extension_amount: int
) -> (list, list):
    """
    Applies the DOS header extension to a sample contained inside a list

    Parameters
    ----------
    x : list
            the sample as a torch Tensor
    preferable_extension_amount : int
            how much extension

    Returns
    -------
    list, list
            returns the perturbed sample and which are the indexes that can be perturbed
    """
    if preferable_extension_amount == 0:
        return x, []
    adv_x = x.flatten().tolist()
    liefpe = lief.PE.parse(adv_x)
    section_file_alignment = liefpe.optional_header.file_alignment
    if section_file_alignment == 0:
        return x, []
    first_content_offset = liefpe.dos_header.addressof_new_exeheader
    extension_amount = (
            int(math.ceil(preferable_extension_amount / section_file_alignment))
            * section_file_alignment
    )
    index_to_perturb = list(range(2, 0x3C)) + list(
        range(0x40, first_content_offset + extension_amount)
    )
    adv_x = _shift_pe_header(liefpe, x, extension_amount)
    for i, _ in enumerate(liefpe.sections):
        adv_x = _shift_pointer_to_section_content(
            liefpe, bytearray(adv_x), i, extension_amount, extension_amount
        )
    x = torch.Tensor(adv_x)
    return x, index_to_perturb


def padding_manipulation(x: torch.Tensor, padding: int) -> (torch.Tensor, list):
    """
    Applies the padding to all vectors in x, returning also the indexes to perturb
    """
    index_to_perturb = list(range(x.shape[-1], x.shape[-1] + padding))
    x = torch.hstack((x, torch.zeros([x.shape[0], padding])))
    return x, index_to_perturb


def content_shift_manipulation(
        x: torch.Tensor, preferable_extension_amount: int, pe_shifted_by: int = 0
) -> (torch.Tensor, list):
    """
    Applies the content shifting to a sample contained inside a list

    Parameters
    ----------
    x : list
            single sample as a torch tensor
    preferable_extension_amount : int
            how much extension
    pe_shifted_by : int, optional, default 0
            if the PE header was shifted, this value should be set to that amount

    Returns
    -------
    list, list
            returns the perturbed sample and which are the indexes that can be perturbed
    """
    if not preferable_extension_amount:
        return x, []
    adv_x = x.flatten().tolist()
    liefpe = lief.PE.parse(adv_x)
    section_file_alignment = liefpe.optional_header.file_alignment
    if section_file_alignment == 0:
        return x, []
    first_content_offset = liefpe.sections[0].offset
    extension_amount = (
            int(math.ceil(preferable_extension_amount / section_file_alignment))
            * section_file_alignment
    )
    index_to_perturb = list(
        range(first_content_offset, first_content_offset + extension_amount)
    )
    for i, _ in enumerate(liefpe.sections):
        adv_x = _shift_pointer_to_section_content(
            liefpe, bytearray(adv_x), i, extension_amount, pe_shifted_by
        )
    adv_x = (
            adv_x[:first_content_offset]
            + b"\x00" * extension_amount
            + adv_x[first_content_offset:]
    )
    x = torch.Tensor(adv_x)
    return x, index_to_perturb


def header_fields_manipulations(pe_index: int):
    # COFF manipulations
    # 4 bytes timestamp at PE + 4 (PE size) + 4
    # 4 bytes PointerToSymbolTable at PE + 4 (PE size) + 8
    # 4 bytes NumberOFSymbols at PE + 4 (PE size) + 12
    #
    # Optional header manipulations
    # 1 byte MajorLinkerVersion at PE + 4 + 20 (COFF size) + 2
    # 1 byte MinorLinkerVersion at PE + 4 + 20 (COFF size) + 3
    # 2 byte MajorOperatingSystemVersion at PE + 4 + 20 (COFF size) + 40
    # 2 byte MinorOperatingSystemVersion at PE + 4 + 20 (COFF size) + 42
    # 2 byte MajorImageVersion at PE + 4 + 20 (COFF size) + 44
    # 2 byte MinorImageVersion at PE + 4 + 20 (COFF size) + 46

    coff_indexes = list(range(pe_index + 4 + 4, pe_index + 4 + 12))
    optional_header_indexes = [pe_index + 24 + 2, pe_index + 24 + 2]
    optional_header_indexes.extend(list(range(pe_index + 24 + 40, pe_index + 24 + 48)))
    indexes = coff_indexes + optional_header_indexes
    return indexes


def section_injection_manipulation(
        x: torch.Tensor, how_many_sections: int, size_per_section: int = 0x200
) -> (list, list):
    pe = lief.PE.parse(x.flatten().tolist())
    default_sect_name = "MLADVEXE"
    for _ in range(how_many_sections):
        section = lief.PE.Section(default_sect_name)
        section.content = [0] * size_per_section
        pe.add_section(section)
    builder = lief.PE.Builder(pe)
    builder.build()
    x_init = builder.get_build()
    pe = lief.PE.parse(raw=x_init)
    index_to_perturb = []
    for section in pe.sections:
        if section.name != default_sect_name:
            continue
        indexes = []  # add 8 of name + index of content
        index_to_perturb.extend(indexes)
    x_init = torch.Tensor(x_init)
    return x_init, index_to_perturb



import torch
from maltorch.datasets.binary_dataset import BinaryDataset


class RandomizedDeletionDataset(BinaryDataset):
    """
    Zhougun Huang, Niel Marchant, Keane Lucas, Lujo Bauer, Olya Ohrimenko, Benjamin I. P. Rubenstein
    RS-Del: Edit Distance Robustness Certificates for Sequence Classifiers via Randomized Deletion
    NeurIPS 2023,
    """
    def __init__(self,
                 csv_filepath: str = None,
                 goodware_directory: str = None,
                 malware_directory: str = None,
                 max_len: int = 2**20,
                 padding_idx: int = 256,
                 min_len: int = None,
                 num_versions: int = 100,
                 pdel: float = 0.97,
                 is_training: bool = True):
        super().__init__(
            csv_filepath=csv_filepath,
            goodware_directory=goodware_directory,
            malware_directory=malware_directory,
            max_len=max_len,
            padding_idx=padding_idx,
            min_len=min_len
        )
        self.num_versions = num_versions
        self.pdel = pdel
        self.is_training = is_training

    def pad_collate_func(self, batch):
        """
        This function randomly deletes the bytes given a probability pdel.
        It works differently at training and at test time.
        During training, given an input example we generate a single randomized version of that example.
        During testing, given an input example we generate N randomized versions of that example.
        """
        vecs = []
        labels = []
        if self.is_training is True:
            for x, y in batch:
                # Get mask
                mask = torch.rand(x.shape[0]) > self.pdel
                masked_x = torch.masked_select(x, mask=mask)
                vecs.append(masked_x)
                labels.append(y)
            x = torch.nn.utils.rnn.pad_sequence(vecs, batch_first=True, padding_value=self.padding_idx)
            # stack will give us (B, 1), so index [:,0] to get to just (B)
            y = torch.tensor(labels)
            return x, y
        else: # Set self.is_training to False whenever you want to generate N versions for each executable.
            if len(batch) == 1:  # Only implemented for batch sizes equals to 1
                x = batch[0][0]
                for i in range(self.num_versions):
                    mask = torch.rand(x.shape[0]) > self.pdel
                    masked_x = torch.masked_select(x, mask=mask)
                    vecs.append(masked_x)
                x = torch.nn.utils.rnn.pad_sequence(vecs, batch_first=True, padding_value=self.padding_idx)
                y = batch[0][1]
                return x, y
            else:
                raise NotImplementedError




import pathlib

from setuptools import find_packages, setup

here = pathlib.Path.cwd()
readme_path = here / "README.md"
version_path = here / "src" / "maltorch" / "VERSION"

CLASSIFIERS = """\
Development Status :: 3 - Alpha
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.8
Programming Language :: Python :: 3.9
Programming Language :: Python :: 3.10
Programming Language :: Python :: Implementation :: PyPy
Topic :: Software Development
Topic :: Scientific/Engineering
"""

# Get the long description from the README file
with readme_path.open() as f:
    long_description = f.read()

# Get the version from VERSION file
with version_path.open() as f:
    version = f.read()

setup(
    name='maltorch',
    version=version,
    packages=find_packages(
        where="src",
        exclude=[
            "*.tests",
            "*.tests.*",
            "tests.*",
            "tests",
        ],
    ),
    package_dir={'': 'src'},
    classifiers=[_f for _f in CLASSIFIERS.split("\n") if _f],
    data_files=[("src/maltorch/VERSION", ["src/maltorch/VERSION"])],
    url='',
    license='MIT',
    author='Luca Demetrio, Daniel Gibert, Andrea Ponte, Maura Pintor',
    author_email='luca.demetrio@unige.it',
    description='Pytorch-based library for creating Adversarial EXEmples against Windows Malware detectors.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        'torch',
        'torchvision',
        'secml-torch',
        'lightgbm',
        'lief',
        'nevergrad',
        'joblib>=1.3.2'
    ],
)



from torch.utils.data import Dataset
from typing import Tuple
import torch
import os
from pathlib import Path


class BinaryDataset(Dataset):
    def __init__(self,
                 csv_filepath: str = None,
                 goodware_directory: str = None,
                 malware_directory: str = None,
                 max_len: int = None,
                 padding_idx: int = 256,
                 min_len: int = None):
        self.all_files = []
        self.max_len = max_len
        self.min_len = min_len
        self.padding_idx = padding_idx

        if csv_filepath is not None:
            with open(csv_filepath, "r") as input_file:
                lines = input_file.readlines()
                for line in lines:
                    tokens = line.strip().split(",")
                    self.all_files.append([tokens[0], int(tokens[1]), os.path.getsize(tokens[0])])
        elif goodware_directory is not None and malware_directory is not None:
            self.all_files.extend(
                [[os.path.join(goodware_directory, filename), 0, os.path.getsize(os.path.join(goodware_directory, filename))] for filename in os.listdir(goodware_directory)])
            self.all_files.extend(
                [[os.path.join(malware_directory, filename), 1, os.path.getsize(os.path.join(malware_directory, filename))] for filename in os.listdir(malware_directory)])
        else:
            raise NotImplementedError("You need to either provide CSV file containing (sample,label) "
                                      "or the paths where the goodware and malware are stored.")

    def __len__(self) -> int:
        return len(self.all_files)

    def __getitem__(self, index) -> Tuple[torch.Tensor, torch.Tensor]:
        to_load, label, _ = self.all_files[index]
        x = load_single_exe(to_load, max_len=self.max_len, min_len=self.min_len, padding_idx=self.padding_idx)
        return x, torch.tensor(label)

    def pad_collate_func(self, batch):
        """
        This should be used as the collate_fn=pad_collate_func for a pytorch DataLoader object in order to pad out files in a batch to the length of the longest item in the batch.
        """
        vecs = [x[0] for x in batch]
        labels = [x[1] for x in batch]

        x = torch.nn.utils.rnn.pad_sequence(vecs, batch_first=True, padding_value=self.padding_idx)
        # stack will give us (B, 1), so index [:,0] to get to just (B)
        y = torch.stack(labels)

        return x, y


def load_single_exe(path: Path, max_len: int = None, min_len: int = None, padding_idx: int = 256) -> torch.Tensor:
    """
    Create a torch.Tensor from the file pointed in the path
    :param path: a pathlib Path
    :return: torch.Tensor containing the bytes of the file as a tensor
    """
    with open(path, "rb") as h:
        if max_len is None:
            code = h.read()
        else:
            code = h.read(max_len)
    x = torch.frombuffer(bytearray(code), dtype=torch.uint8)
    x = x.to(torch.long)
    if min_len is not None: # Pad the tensor to the minimum length - required for some architectures
        if x.shape[0] < min_len:
            padding_idx = torch.tensor(padding_idx, dtype=torch.long)
            x = torch.nn.functional.pad(x, (0, min_len - x.shape[0]), mode='constant', value=padding_idx)
    return x



from typing import Callable, Optional

import torch.optim
from torch import Tensor

from maltorch.optim.base import BaseByteOptimizer

INVALID = torch.inf


class BGD(BaseByteOptimizer):
    def step(self, closure: Optional[Callable[[], float]] = ...) -> Optional[float]:
        for group in self.param_groups:
            for delta in group["params"]:
                if self._embedding_grad is None:
                    raise ValueError(
                        "Grad is none, you should call backward first before invoking the step function. "
                        "Otherwise, the model itself is not differentiable and gradients can not be computed."
                    )
                grad_embedding = self._embedding_grad[0]
                grad_embedding = self.gradient_processing(grad_embedding)
                updated_delta = token_gradient_descent(
                    embedding_tokens=self.embedding_matrix,
                    emb_x=self.embedding_matrix[delta.long()],
                    gradient_f=grad_embedding,
                    step_size=self.step_size,
                    index_to_perturb=self.indexes_to_perturb,
                    admitted_tokens=torch.LongTensor(range(0, 256)),
                    unavailable_tokens=torch.LongTensor([256]),
                )
                updated_delta = updated_delta.to(delta.device)
                to_update = updated_delta != INVALID

                if torch.any(to_update):
                    delta.data[to_update] = updated_delta[to_update]
        return None


def token_gradient_descent(
    embedding_tokens: torch.Tensor,
    emb_x: torch.Tensor,
    gradient_f: torch.Tensor,
    step_size: int,
    index_to_perturb: torch.LongTensor,
    admitted_tokens: torch.LongTensor,
    unavailable_tokens: Optional[torch.Tensor] = None,
) -> Tensor:
    optimized_tokens = torch.zeros(*emb_x.shape[:2]) + INVALID
    for j in range(emb_x.shape[0]):
        step_size_index = (
            gradient_f[j, :][index_to_perturb[j].view(-1)]
            .norm(dim=1)
            .argsort()[:step_size]
        )
        for i in step_size_index:
            gradient_f_i = -gradient_f[j, i]
            x_i = emb_x[j, i]
            token_to_chose = single_token_gradient_update(
                start_token=x_i.cpu(),
                gradient=gradient_f_i.cpu(),
                embedded_tokens=embedding_tokens.cpu(),
                admitted_tokens=admitted_tokens.cpu(),
                unavailable_tokens=unavailable_tokens,
            )
            optimized_tokens[j, i] = token_to_chose
    return optimized_tokens


def single_token_gradient_update(
    start_token: torch.Tensor,
    gradient: torch.Tensor,
    embedded_tokens: torch.Tensor,
    admitted_tokens: torch.LongTensor,
    invalid_val=INVALID,
    unavailable_tokens: Optional[torch.Tensor] = None,
):
    """
    Given the starting byte, the gradient and the embedding map,it returns a list of distances

    Parameters
    ----------
    start_token : torch.Tensor
        the starting embedding token for the search
    gradient : torch.Tensor
        the gradient of a single embedded token
    embedded_tokens : torch.Tensor
        the embedding matrix with all the byte embedded
    admitted_tokens: list
        the list of indexes of the tokens to use in the search
    invalid_val : optional, default torch.inf
        the invalid value to use. Default torch.inf
    unavailable_tokens: Union[torch.Tensor, None] = None
        if specified, it avoids the usage of the selected tokens during the search step
    Returns
    -------

    """
    if torch.equal(gradient, torch.zeros(gradient.shape)):
        return INVALID
    distance = torch.zeros(len(admitted_tokens))
    for i, b in enumerate(embedded_tokens[admitted_tokens, :].cpu()):
        if torch.all(start_token == b):
            distance[i] = invalid_val
            continue
        if unavailable_tokens is not None:
            if i in unavailable_tokens:
                distance[i] = INVALID
        bts = b - start_token
        s_i = torch.dot(gradient, bts)
        if s_i <= 0:
            distance[i] = invalid_val
        else:
            d_i = torch.norm(b - (start_token + s_i * gradient))
            distance[i] = d_i
    min_value, token_to_chose = torch.min(distance, dim=0, keepdim=True)
    if min_value == INVALID:
        return INVALID
    return token_to_chose



"""
Nebula: Self-Attention for Dynamic Malware Analysis
Dmitrijs Trizna; Luca Demetrio; Battista Biggio; Fabio Roli
IEEE Transactions on Information Forensics andSecurity (TIFS), 2024
https://ieeexplore.ieee.org/document/10551436
Reimplemented from: https://github.com/dtrizna/nebula/blob/main/nebula/__init__.py
"""

import math
import importlib.util
from typing import Union, Optional, List

import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from torch.nn import TransformerEncoder, TransformerEncoderLayer

from maltorch.zoo.model import EmbeddingModel


class PositionalEncoding(nn.Module):
    """ From: https://pytorch.org/tutorials/beginner/transformer_tutorial.html """

    def __init__(self, d_model: int, dropout: float = 0.1, max_len: int = 5000):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = torch.zeros(max_len, 1, d_model)
        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2) * (-math.log(10000.0) / d_model))
        pe[:, 0, 0::2] = torch.sin(position * div_term)
        pe[:, 0, 1::2] = torch.cos(position * div_term)
        self.register_buffer('pe', pe)

    def forward(self, x: Tensor) -> Tensor:
        """
        Args:
            x: Tensor, shape [seq_len, batch_size, embedding_dim]
        """
        x = x + self.pe[:x.size(0)]
        return self.dropout(x)


class TransformerEncoderModel(nn.Module):
    def __init__(
            self,
            vocab_size: int,  # size of vocabulary
            maxlen: int,  # maximum length of input sequence
            dModel: int = 32,  # embedding & transformer dimension
            nHeads: int = 8,  # number of heads in nn.MultiheadAttention
            dHidden: int = 200,  # dimension of the feedforward network model in nn.TransformerEncoder
            nLayers: int = 2,  # number of nn.TransformerEncoderLayer in nn.TransformerEncoder
            numClasses: int = 1,  # 1 ==> binary classification
            classifier_head: Optional[list] = [64],  # decoder's classifier FFNN complexity
            pretrain_layers: Optional[List] = None,
            layerNorm: bool = False,  # whether to normalize decoder's FFNN layers
            norm_first: bool = True,  # whether to normalize before or after FFNN layers
            dropout: float = 0.3,
            pooling: str = "mean",
            skip_embedding: bool = False,
            causal_attention: bool = False
    ):
        super().__init__()
        self.__name__ = 'TransformerEncoderModel'
        
        assert dModel % nHeads == 0, "nheads must divide evenly into d_model"
        assert pooling in ["mean", "flatten", "cls", None]

        self.vocab_size = vocab_size
        self.maxlen = maxlen
        self.skip_embedding = skip_embedding
        self.encoder = nn.Embedding(vocab_size, dModel)
        self.pos_encoder = PositionalEncoding(dModel, dropout)
        encoder_layers = TransformerEncoderLayer(
            d_model=dModel,
            nhead=nHeads,
            dim_feedforward=dHidden,
            dropout=dropout,
            batch_first=True,
            norm_first=norm_first
        )
        self.transformer_encoder = TransformerEncoder(encoder_layers, nLayers)
        self.d_model = dModel
        self.layerNorm = layerNorm
        self.dropout = dropout
        
        self.pooling_type = pooling
        if pooling == "mean" or pooling is None:
            input_neurons = self.d_model
        elif pooling == "flatten":
            input_neurons = int(self.maxlen * dModel)
        elif pooling == "cls":
            raise NotImplementedError
        
        self.causal_attention = causal_attention
        if self.causal_attention: # override pooling settings if causal
            input_neurons = self.d_model
            self.pooling_type = None
        
        self.classifier_head_nr_list = classifier_head
        self.num_classes_out = numClasses
        self.classifier_head = None
        if self.classifier_head_nr_list is not None:
            # model is initiated as a downstream model -- setup classifier head
            self.classifier_head_layers = []
            if len(self.classifier_head_nr_list) > 0:
                self.classifier_head_layers = self._build_ffnn_layers(
                    start_neurons=input_neurons,
                    hidden_layers=self.classifier_head_nr_list
                )
            # if numClasses is None, then classifier outputs last value from classified_head
            if self.num_classes_out is not None:
                self.final_layer_in = classifier_head[-1] if len(classifier_head) > 0 else input_neurons
                self.final_layer_out =  1 if self.num_classes_out == 2 else self.num_classes_out
                self.classifier_head_layers.append(nn.Linear(self.final_layer_in, self.final_layer_out))
            # join in a single classifier head
            self.classifier_head = nn.Sequential(*self.classifier_head_layers)
            
        self.pretrain_layers = None
        if pretrain_layers is not None:
            self.pretrain_layers = []
            if len(pretrain_layers) > 0:
                self.pretrain_layers = self._build_ffnn_layers(
                    start_neurons=input_neurons,
                    hidden_layers=pretrain_layers
                )
            final_layer_in = pretrain_layers[-1] if len(pretrain_layers) > 0 else input_neurons
            # NOTE: bias removed in last layer as in https://github.com/karpathy/nanoGPT/blob/master/model.py#L133
            self.pretrain_layers.append(nn.Linear(final_layer_in, self.vocab_size, bias=False))
            self.pretrain_layers = nn.Sequential(*self.pretrain_layers)

        self.apply(self._init_weights)

    def _init_weights(self, module):
        # NOTE: from: https://github.com/karpathy/nanoGPT/blob/master/model.py#L162
        if isinstance(module, nn.Linear):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
            if module.bias is not None:
                torch.nn.init.zeros_(module.bias)
        elif isinstance(module, nn.Embedding):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)

    def _build_ffnn_layers(self, start_neurons: int, hidden_layers: list) -> nn.Sequential:
        ffnn = []
        if hidden_layers is None:
            return ffnn
        for i, h in enumerate(hidden_layers):
            ffnnBlock = []
            if i == 0:
                ffnnBlock.append(nn.Linear(start_neurons, h))
            else:
                ffnnBlock.append(nn.Linear(hidden_layers[i - 1], h))

            # add LayerNorm to every layer except last
            if self.layerNorm and i < len(hidden_layers) - 1:
                ffnnBlock.append(nn.LayerNorm(h))

            ffnnBlock.append(nn.ReLU())
            ffnnBlock.append(nn.Dropout(self.dropout))

            ffnn.append(nn.Sequential(*ffnnBlock))
        return ffnn
    
    @staticmethod
    def _generate_square_subsequent_mask(sz: int) -> Tensor:
        mask = (torch.triu(torch.ones(sz, sz)) == 1).transpose(0, 1)
        mask = mask.float().masked_fill(mask == 0, float('-inf')).masked_fill(mask == 1, float(0.0))
        return mask

    def embed(self, x: Tensor) -> Tensor:
        encoded_x = self.encoder(x) * math.sqrt(self.d_model)
        encoded = self.pos_encoder(encoded_x)
        return encoded

    def pooling(self, x: Tensor) -> Tensor:
        if self.causal_attention or self.pooling_type is None:
            return x
        if self.pooling_type == "flatten":
            x = x.view(x.size(0), -1)
        if self.pooling_type == "mean":
            x = torch.mean(x, dim=1)
        return x

    def core(self, src: Tensor, src_mask: Optional[Tensor] = None) -> Tensor:
        x = src if self.skip_embedding else self.embed(src)
        if self.causal_attention:
            src_mask = self._generate_square_subsequent_mask(x.shape[1]).to(x.device)
        x = self.transformer_encoder(x, src_mask, is_causal=self.causal_attention)
        x = self.pooling(x)
        return x

    def forward(self, x: Tensor, src_mask: Optional[Tensor] = None) -> Tensor:
        x = self.core(x, src_mask)
        if self.classifier_head is not None:
            x = self.classifier_head(x)
        return x

    def pretrain(self, x: Tensor) -> Tensor:
        x = self.core(x)
        if self.pretrain_layers is not None:
            x = self.pretrain_layers(x)
        return x


class TransformerEncoderChunks(TransformerEncoderModel):
    """
    Slices global attention to multiple separate chunks, which are processed independently till classifier head.
    Main model used in: https://arxiv.org/abs/2310.10664
    'Chunk' term borrowed from longformer:
    https://github.com/allenai/longformer/blob/master/longformer/sliding_chunks.py
    """

    def __init__(self, chunk_size: int = 64, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__name__ = 'TransformerEncoderChunks'

        self.chunk_size = chunk_size
        self.nr_of_chunks = self.maxlen / self.chunk_size
        if self.nr_of_chunks != int(self.nr_of_chunks):
            self.nr_of_chunks = int(self.nr_of_chunks) + 1
        
    def split(self, src: Tensor) -> List[Tensor]:
        chunks = []
        for chunk in torch.split(src, split_size_or_sections=self.chunk_size, dim=1):
            if chunk.shape[1] < self.chunk_size:
                pad_mask = (0, self.chunk_size - chunk.shape[1])
                chunk = F.pad(chunk, pad=pad_mask)
            chunks.append(chunk)
        return chunks

    def embed(self, chunks: List[Tensor]) -> List[Tensor]:
        encoded_chunks = []
        for chunk in chunks:
            encoded_chunk = self.encoder(chunk) * math.sqrt(self.d_model)
            encoded_chunk = self.pos_encoder(encoded_chunk)
            encoded_chunks.append(encoded_chunk)
        return encoded_chunks

    def transform(self, chunks: List[Tensor], src_mask: Optional[Tensor] = None) -> List[Tensor]:
        transformed_chunks = []
        for chunk in chunks:
            transformed_chunk = self.transformer_encoder(chunk, src_mask)
            transformed_chunks.append(transformed_chunk)
        return transformed_chunks

    def core(self, src: Tensor, src_mask: Optional[Tensor] = None) -> Tensor:
        chunks = self.split(src)
        chunks = self.embed(chunks)  # [(batch_size, chunk_size, d_model), ..]
        chunks = self.transform(chunks, src_mask)  # [(batch_size, chunk_size, d_model), ..]
        x = torch.cat(chunks, dim=1)
        # NOTE: after .cat() shape is: (batch_size, nr_of_chunks * chunk_size, d_model)
        # where nr_of_chunks = int(maxlen / self.chunk_size) + 1
        x = self.pooling(x)
        return x


class Nebula(EmbeddingModel):
    def __init__(
            self,
            vocab_size: int = 50000,
            seq_len: int = 512,
            tokenizer: str = 'whitespace',
    ):
        # optional library imports
        speakeasy_spec = importlib.util.find_spec("speakeasy")
        if speakeasy_spec is None:
            raise ImportError("[-] 'speakeasy' is required for emulation module. Use 'pip install speakeasy' to install.")
        from maltorch.data_processing.dynamic.emulation import PEDynamicFeatureExtractor
        
        nltk_spec = importlib.util.find_spec("nltk")
        if nltk_spec is None:
            raise ImportError("[-] 'nltk' is required for tokenization module. Use 'pip install nltk' to install.")
        from maltorch.data_processing.dynamic.tokenization import JSONTokenizerNaive, JSONTokenizerBPE

        self.vocab_size = vocab_size
        self.seq_len = seq_len

        # dynamic extractor setup
        self.dynamic_extractor = PEDynamicFeatureExtractor() 
        
        # tokenizer initialization
        assert tokenizer in ['bpe', 'whitespace'], "tokenizer must be in ['bpe', 'whitespace']"
        if tokenizer == 'bpe':
            self.tokenizer = JSONTokenizerBPE(
                vocab_size=self.vocab_size,
                seq_len=self.seq_len,
            )
        if tokenizer == 'whitespace':
            self.tokenizer = JSONTokenizerNaive(
                vocab_size=self.vocab_size,
                seq_len=self.seq_len,
            )
        
        # PyTorch model initialization
        torch_model_config = {
            "vocab_size": self.vocab_size,
            "maxlen": self.seq_len,
            "chunk_size": 64, # self-attention window size
            "dModel": 64,  # embedding & transformer dimension
            "nHeads": 8,  # number of heads in nn.MultiheadAttention
            "dHidden": 256,  # dimension of the feedforward network model in nn.TransformerEncoder
            "nLayers": 2,  # number of nn.TransformerEncoderLayer in nn.TransformerEncoder
            "numClasses": 1, # binary classification
            "classifier_head": [64], # classifier head depth
            "layerNorm": False,
            "dropout": 0.3,
            "norm_first": True,
            "pooling": "flatten"
        }
        self.model = TransformerEncoderChunks(**torch_model_config)

    def dynamic_analysis_pe_file(self, pe_file: Union[str, bytes]) -> dict:
        if isinstance(pe_file, str):
            with open(pe_file, "rb") as f:
                bytez = f.read()
        elif isinstance(pe_file, bytes):
            bytez = pe_file
        else:
            raise ValueError("preprocess(): data must be a path to a PE file or a bytes object")

        dynamic_features_json = self.dynamic_extractor.emulate(data=bytez)
        return dynamic_features_json

    def preprocess(self, emulation_report: dict) -> Tensor:
        dynamic_features = self.tokenizer.encode(emulation_report)
        return torch.Tensor(dynamic_features).long()
    
    def predict_proba(self, dynamic_features: Tensor) -> float:
        dynamic_features = torch.Tensor(dynamic_features).long()
        with torch.no_grad():
            logits = self.model(dynamic_features)
        return torch.sigmoid(logits).item()
    
    def predict_sample(self, pe_file: Union[str, bytes]) -> float:
        dynamic_features_json = self.dynamic_analysis_pe_file(pe_file)
        dynamic_features = self.preprocess(dynamic_features_json)
        return self.predict_proba(dynamic_features)

    def embed(self, x: Tensor) -> Tensor:
        return self.model.embed(x)
    
    def embedding_layer(self):
        return self.model.encoder
    
    def embedding_matrix(self):
        return self.model.encoder.weight
    
    def forward(self, x: Tensor) -> Tensor:
        return self.model(x)



from abc import ABC, abstractmethod
from typing import Optional, Union

import torch
from secmlt.models.base_model import BaseModel
from secmlt.models.base_trainer import BaseTrainer
from secmlt.models.data_processing.data_processing import DataProcessing
from secmlt.models.pytorch.base_pytorch_nn import BasePytorchClassifier

from maltorch.utils.config import Config
from maltorch.utils.utils import download_gdrive


class Model(torch.nn.Module, ABC):
    def __init__(self, name: str, gdrive_id: Optional[str]):
        super().__init__()
        self.name = name
        self.gdrive_id = gdrive_id
        self.model_path = Config.MODEL_ZOO_FOLDER / self.name

    def _fetch_pretrained_model(self):
        if not Config.MODEL_ZOO_FOLDER.exists():
            Config.MODEL_ZOO_FOLDER.mkdir()
        if not self.model_path.exists():
            if self.gdrive_id is not None:
                download_gdrive(gdrive_id=self.gdrive_id, fname_save=self.model_path)

    def load_pretrained_model(self, device="cpu", model_path=None): ...

    @classmethod
    def create_model(
        cls,
        model_path: Optional[str] = None,
        device: str = "cpu",
        preprocessing: DataProcessing = None,
        postprocessing: DataProcessing = None,
        trainer: BaseTrainer = None,
    ) -> BaseModel: ...


class PytorchModel(Model):
    def load_pretrained_model(self, device="cpu", model_path=None):
        if self.gdrive_id is None:
            return
        path = self.model_path
        if model_path is None:
            self._fetch_pretrained_model()
        else:
            path = model_path
        state_dict = torch.load(path, map_location=device, weights_only=True)
        self.load_state_dict(state_dict)

    @classmethod
    def create_model(
        cls,
        model_path: Optional[str] = None,
        device: str = "cpu",
        preprocessing: DataProcessing = None,
        postprocessing: DataProcessing = None,
        trainer: BaseTrainer = None,
        **kwargs,
    ) -> BaseModel:
        net = cls(**kwargs)
        net.load_pretrained_model(device=device, model_path=model_path)
        net = net.to(device) # Explicitly load model to device
        net.eval()
        net = BasePytorchClassifier(
            model=net,
            preprocessing=preprocessing,
            postprocessing=postprocessing,
            trainer=trainer,
        )
        return net


class BaseEmbeddingPytorchClassifier(BasePytorchClassifier):
    def __init__(
        self,
        model: torch.nn.Module,
        preprocessing: DataProcessing = None,
        postprocessing: DataProcessing = None,
        trainer: BaseTrainer = None,
        threshold: Optional[Union[float, None]] = 0.5,
    ):
        super().__init__(model, preprocessing, postprocessing, trainer)
        self.threshold = threshold

    def embed(self, x: torch.Tensor):
        return self.model.embed(x)

    def embedding_matrix(self):
        return self.model.embedding_matrix()

    def embedding_layer(self):
        return self.model.embedding_layer()

    def predict(self, x: torch.Tensor):
        if self.threshold is None:
            return super().predict(x)
        scores = self.decision_function(x)
        labels = (scores > self.threshold).int()
        return labels


class EmbeddingModel(PytorchModel, ABC):
    @classmethod
    def create_model(
        cls,
        model_path: Optional[str] = None,
        device: str = "cpu",
        preprocessing: DataProcessing = None,
        postprocessing: DataProcessing = None,
        trainer: BaseTrainer = None,
        threshold: Optional[Union[float, None]] = 0.5,
        **kwargs,
    ) -> BaseEmbeddingPytorchClassifier:
        net = cls(**kwargs)
        net.load_pretrained_model(device=device, model_path=model_path)
        net = net.to(device) # Explicitly load model to device
        net.eval()
        net = BaseEmbeddingPytorchClassifier(
            model=net,
            preprocessing=preprocessing,
            postprocessing=postprocessing,
            trainer=trainer,
            threshold=threshold,
        )
        return net

    def __init__(
        self, name: str, gdrive_id: Optional[str], input_embedding: bool = False
    ):
        super().__init__(name, gdrive_id)
        self.input_embedding = input_embedding

    @abstractmethod
    def embed(self, x):
        pass

    @abstractmethod
    def embedding_layer(self):
        pass

    @abstractmethod
    def embedding_matrix(self):
        pass

    @abstractmethod
    def _forward_embed_x(self, x):
        pass

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass.

        Parameters
        ----------
        x : torch.Tensor
            the sample to test
        Returns
        -------
        torch.Tensor
            the result of the forward pass
        """
        x = self.embed(x)
        output = self._forward_embed_x(x)
        return output


class BaseGrayscalePytorchClassifier(BasePytorchClassifier):
    def __init__(
        self,
        model: torch.nn.Module,
        preprocessing: DataProcessing = None,
        postprocessing: DataProcessing = None,
        trainer: BaseTrainer = None,
        threshold: Optional[Union[float, None]] = 0.5,
    ):
        super().__init__(model, preprocessing, postprocessing, trainer)
        self.threshold = threshold

    def predict(self, x: torch.Tensor):
        if self.threshold is None:
            return super().predict(x)
        scores = self.decision_function(x)
        labels = (scores > self.threshold).int()
        return labels



import torch

from maltorch.initializers.initializers import ByteBasedInitializer
from maltorch.utils.pe_operations import section_injection_manipulation


class SectionInjectionInitializer(ByteBasedInitializer):
    def __init__(
        self,
        how_many_sections: int,
        size_per_section: int = 0x200,
        random_init: bool = False,
    ):
        super().__init__(random_init=random_init)
        self.how_many_sections = how_many_sections
        self.size_per_section = size_per_section

    def __call__(self, x: torch.Tensor) -> [torch.Tensor, torch.Tensor]:
        X = []
        deltas = []
        indexes = []
        for x_i in x:
            x_i, shift_indexes = section_injection_manipulation(
                x_i, self.how_many_sections, self.size_per_section
            )
            X.append(x_i)
            delta = (
                torch.zeros(len(shift_indexes))
                if not self.random_init
                else torch.randint(0, 255, (len(shift_indexes),))
            )
            deltas.append(delta)
            indexes.append(shift_indexes)
        x, delta, indexes = self._pad_samples_same_length(X, deltas, indexes)
        return x, delta, indexes



from functools import partial
from typing import Union, Callable

from nevergrad.optimization.differentialevolution import (
    DifferentialEvolution,
)

from maltorch.optim.bgd import BGD
from maltorch.optim.byte_gradient_processing import ByteGradientProcessing


class MalwareOptimizerFactory:
    @staticmethod
    def create(optim_cls: Union[str, Callable], **optimizer_args):
        if not isinstance(optim_cls, str):
            return partial(optim_cls, **optimizer_args)()
        if optim_cls == "bgd":
            return MalwareOptimizerFactory.create_bgd(**optimizer_args)
        if optim_cls == "ga":
            return MalwareOptimizerFactory.create_ga()
        raise NotImplementedError(f"Optimizer {optim_cls} not included.")

    @staticmethod
    def create_bgd(lr: int, device: str = "cpu") -> partial[BGD]:
        return partial(
            BGD,
            lr=lr,
            gradient_processing=ByteGradientProcessing(),
            device=device,
        )

    @staticmethod
    def create_ga(population_size: int = 10) -> DifferentialEvolution:
        return DifferentialEvolution(popsize=population_size, crossover="twopoints")



import torch
from secmlt.models.data_processing.data_processing import DataProcessing
import math


class SequentialDeRandomizedPreprocessing(DataProcessing):
    """
    Daniel Gibert, Giulio Zizzo, Quan Le, Jordi Planes
    Adversarial Robustness of Deep Learning-based Malware Detectors via (De) Randomized Smoothing
    IEEE Access 2024 - Sequential Chunks-based (De)Randomized Smoothing
    """
    def __init__(self,
                 file_percentage: float = 0.05,
                 num_chunks: int = 100,
                 padding_idx: int = 256,
                 min_chunk_size: int = 500):
        super().__init__()
        self.file_percentage = file_percentage
        self.num_chunks = num_chunks
        self.padding_idx = padding_idx
        self.min_chunk_size = min_chunk_size

    def _process(self, x: torch.Tensor) -> torch.Tensor:
        x = x.squeeze()  # Remove all dimensions equal to 1
        max_size = x.size()[0]
        group_size = math.ceil(max_size * self.file_percentage)
        group_size = max(self.min_chunk_size,
                         group_size)  # The chunk size has to be at least equal to the kernel size of the first convolutional layer
        overlap_size = group_size - int(max_size / self.num_chunks)

        to_substract_A = math.ceil((group_size - overlap_size) / self.num_chunks)
        to_substract_B = math.ceil((group_size - overlap_size) * self.file_percentage)
        to_substract = to_substract_B - to_substract_A

        vecs = []
        for i in range(self.num_chunks):
            start = i * (group_size - overlap_size - to_substract)
            end = start + group_size
            vecs.append(x[start:end])

        x = torch.nn.utils.rnn.pad_sequence(vecs, batch_first=True, padding_value=self.padding_idx)
        return x

    def invert(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the inverted transform (if defined).

        Parameters
        ----------
        x : torch.Tensor
            Input samples.

        Returns
        -------
        torch.Tensor
            The samples in the input space before the transformation.
        """
        raise NotImplementedError




from typing import Tuple

import torch
from maltorch.datasets.dynamic_drs_dataset import DynamicChunkSizeDRSDataset
import math


class SequentialDRSDataset(DynamicChunkSizeDRSDataset):
    """
    Daniel Gibert, Giulio Zizzo, Quan Le, Jordi Planes
    Adversarial Robustness of Deep Learning-based Malware Detectors via (De) Randomized Smoothing
    IEEE Access 2024 -Sequential Chunks-based (De)Randomized Smoothing
    """
    def __init__(self,
                 csv_filepath: str = None,
                 goodware_directory: str = None,
                 malware_directory: str = None,
                 max_len: int = 2 ** 20,
                 padding_idx: int = 256,
                 min_len: int = None,
                 sort_by_size: bool = False,
                 file_percentage: float = 0.05,
                 num_chunks: int = 100,
                 is_training: bool = True,
                 min_chunk_size=500):
        super().__init__(
            csv_filepath=csv_filepath,
            goodware_directory=goodware_directory,
            malware_directory=malware_directory,
            max_len=max_len,
            padding_idx=padding_idx,
            min_len=min_len,
            sort_by_size=sort_by_size,
            file_percentage=file_percentage,
            num_chunks=num_chunks,
            is_training=is_training,
            min_chunk_size=min_chunk_size
        )

    def generate_testing_examples(self, batch) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        During testing, given an input example we sequentially extract N chunks of bytes. This chunks may overlap with
        one another depending on the amount of chunks you want to extract and the size of each chunk
        """

        if len(batch) == 1:  # Only implemented for batch sizes equals to 1
            x = batch[0][0]
            max_size = x.size()[0]

            group_size = math.ceil(max_size * self.file_percentage)
            group_size = max(self.min_chunk_size, group_size) # The chunk size has to be at least equal to the kernel size of the first convolutional layer
            overlap_size = group_size - int(max_size / self.num_chunks)

            to_substract_A = math.ceil((group_size - overlap_size) / self.num_chunks)
            to_substract_B = math.ceil((group_size - overlap_size) * self.file_percentage)
            to_substract = to_substract_B - to_substract_A

            vecs = []
            for i in range(self.num_chunks):
                start = i * (group_size - overlap_size - to_substract)
                end = start + group_size
                vecs.append(x[start:end])

            x = torch.nn.utils.rnn.pad_sequence(vecs, batch_first=True, padding_value=self.padding_idx)
            y = batch[0][1]
            return x, y
        else:
            raise NotImplementedError




from typing import Optional

import ember.features
import lightgbm
import numpy as np
import torch
from secmlt.models.base_model import BaseModel
from secmlt.models.base_trainer import BaseTrainer
from secmlt.models.data_processing.data_processing import DataProcessing
from torch.utils.data import DataLoader

from maltorch.zoo.model import Model


class EmberGBDT(Model):
    def __init__(self, model_path: Optional[str] = None):
        super().__init__(
            name="ember_gbdt", gdrive_id="1RWvr3yD8M90EXcTozK2TwW2JEExQ9qDW"
        )
        self._fetch_pretrained_model()
        self.tree_model = None
        if model_path is None:
            model_path = self.model_path
        self.load_pretrained_model(model_path=model_path)

    def load_pretrained_model(self, device: str = "cpu", model_path: str = None):
        if model_path is None:
            self._fetch_pretrained_model()
            self.tree_model = lightgbm.Booster(model_file=self.model_path)
        else:
            self.tree_model = lightgbm.Booster(model_file=model_path)

    @classmethod
    def create_model(
            cls,
            model_path: Optional[str] = None,
            threshold: int = 0.82,
            device: str = "cpu",
            preprocessing: DataProcessing = None,
            postprocessing: DataProcessing = None,
            trainer: BaseTrainer = None,
    ) -> BaseModel:
        model = cls(model_path=model_path)
        return _GBDTModel(tree_model=model.tree_model, threshold=threshold)


class _GBDTModel(BaseModel):
    def __init__(self, tree_model: lightgbm.Booster, threshold: float = 0.82):
        super().__init__()
        self.tree_model: lightgbm.Booster = tree_model
        self.threshold: float = threshold

    def predict(self, x: torch.Tensor) -> torch.Tensor:
        y = self._decision_function(x)
        return torch.Tensor(y > self.threshold).int()

    def _decision_function(self, x: torch.Tensor) -> torch.Tensor:
        extractor = ember.features.PEFeatureExtractor(print_feature_warning=False)
        n_samples = x.shape[0]
        feat_x = np.zeros((x.shape[0], extractor.dim))
        for i in range(n_samples):
            x_i = x[i, :]
            x_bytes = bytes(x_i.type(torch.int).flatten().tolist())
            feat_x[i, :] = torch.Tensor(extractor.feature_vector(x_bytes))
        probabilities = self.tree_model.predict(feat_x)
        probabilities = torch.Tensor(probabilities).unsqueeze(1)
        return probabilities

    def gradient(self, x: torch.Tensor, y: int) -> torch.Tensor:
        raise NotImplementedError("No gradient for GBDT model.")

    def train(self, dataloader: DataLoader) -> "BaseModel":
        raise NotImplementedError("Trainer not implemented yet.")
